from .. import optim
from .. import stats

from . import base


__all__ = ['LogLoss']


class LogLoss(stats.Mean, base.BinaryClassificationMetric):
    """Binary logarithmic loss.

    Example:

        ::

            >>> from creme import metrics
            >>> from sklearn.metrics import log_loss

            >>> y_true = [True, False, False, True]
            >>> y_pred = [0.9,  0.1,   0.2,   0.65]

            >>> metric = metrics.LogLoss()
            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     metric = metric.update(y_t, y_p)
            ...     print(metric.get())
            0.105360...
            0.105360...
            0.144621...
            0.216161...

            >>> metric
            LogLoss: 0.216162

    """

    @property
    def bigger_is_better(self):
        return False

    @property
    def requires_labels(self):
        return False

    def update(self, y_true, y_pred):
        ll = optim.LogLoss().__call__
        return super().update(ll(y_true, y_pred[True] if isinstance(y_pred, dict) else y_pred))



"""
A module for ensemble learning.
"""
from .bagging import BaggingClassifier
from .bagging import BaggingRegressor
from .hedge import HedgeClassifier


__all__ = [
    'BaggingClassifier',
    'BaggingRegressor',
    'HedgeClassifier'
]



import copy
import functools
import itertools
import math
import random

from sklearn import datasets


__all__ = ['check_estimator']


def make_random_features(model, n_observations, n_features):
    for _ in range(n_observations):
        yield {i: random.random() for i in range(n_features)}


def make_random_targets(model, n_observations):

    from .. import base

    random_func = None

    if isinstance(model, base.Regressor):
        random_func = random.random

    elif isinstance(model, base.MultiClassifier):
        random_func = functools.partial(random.choice, ['a', 'b', 'c', 'd'])

    elif isinstance(model, base.BinaryClassifier):
        random_func = functools.partial(random.choice, [True, False])

    elif isinstance(model, base.Transformer):
        if model.is_supervised:
            random_func = random.random
        else:
            random_func = lambda: None

    elif isinstance(model, base.Clusterer):
        random_func = functools.partial(random.choice, [0, 1, 2, 3])

    for _ in range(n_observations):
        yield random_func()


def make_random_X_y(model, n_observations, n_features):
    return zip(
        make_random_features(model, n_observations, n_features),
        make_random_targets(model, n_observations)
    )


def check_fit_one(model):

    klass = model.__class__

    for x, y in make_random_X_y(model, n_observations=20, n_features=4):

        xx, yy = copy.deepcopy(x), copy.deepcopy(y)

        model = model.fit_one(x, y)

        # Check the model returns itself
        assert isinstance(model, klass)

        # Check fit_one is pure (i.e. x and y haven't changed)
        assert x == xx
        assert y == yy


def check_predict_proba_one(model):

    for x, y in make_random_X_y(model, n_observations=20, n_features=4):

        xx, yy = copy.deepcopy(x), copy.deepcopy(y)

        model = model.fit_one(x, y)
        y_pred = model.predict_proba_one(x)

        # Check the probabilities are coherent
        assert isinstance(y_pred, dict)
        assert math.isclose(sum(y_pred.values()), 1.0)
        for label, proba in y_pred.items():
            assert 0.0 <= proba <= 1.0

        # Check predict_proba_one is pure (i.e. x and y haven't changed)
        assert x == xx
        assert y == yy


def check_a_better_than_b(model_a, model_b, X_y_func, metric):
    """Trains two models and checks that model_a does better than model_b."""

    from .. import model_selection

    metric_a = model_selection.online_score(
        X_y=X_y_func(),
        model=model_a,
        metric=copy.deepcopy(metric)
    )
    metric_b = model_selection.online_score(
        X_y=X_y_func(),
        model=model_b,
        metric=copy.deepcopy(metric)
    )

    if metric.bigger_is_better:
        assert metric_a.get() > metric_b.get()
    else:
        assert metric_a.get() < metric_b.get()


def check_better_than_dummy_binary(classifier):

    from .. import dummy
    from .. import metrics
    from .. import stream

    for dummy_model, X_y_func, metric in itertools.product(
        (dummy.NoChangeClassifier(), dummy.PriorClassifier()),
        (functools.partial(
            stream.iter_sklearn_dataset,
            datasets.load_breast_cancer,
            shuffle=True,
            random_state=42,
        ),),
        (metrics.Accuracy(),)
    ):
        check_a_better_than_b(
            model_a=copy.deepcopy(classifier),
            model_b=copy.deepcopy(dummy_model),
            X_y_func=X_y_func,
            metric=copy.deepcopy(metric)
        )


def check_better_than_dummy_multi(classifier):

    from .. import dummy
    from .. import metrics
    from .. import stream

    for dummy_model, X_y_func, metric in itertools.product(
        (dummy.NoChangeClassifier(), dummy.PriorClassifier()),
        (functools.partial(
            stream.iter_sklearn_dataset,
            datasets.load_iris,
            shuffle=True,
            random_state=42,
        ),),
        (metrics.Accuracy(),)
    ):
        check_a_better_than_b(
            model_a=copy.deepcopy(classifier),
            model_b=copy.deepcopy(dummy_model),
            X_y_func=X_y_func,
            metric=copy.deepcopy(metric)
        )


def check_better_than_dummy_regression(regressor):

    from .. import dummy
    from .. import metrics
    from .. import stats
    from .. import stream

    for dummy_model, X_y_func, metric in itertools.product(
        (dummy.StatisticRegressor(stats.Mean()),),
        (functools.partial(
            stream.iter_sklearn_dataset,
            datasets.load_boston,
            shuffle=True,
            random_state=42,
        ),),
        (metrics.MSE(),)
    ):
        check_a_better_than_b(
            model_a=copy.deepcopy(regressor),
            model_b=copy.deepcopy(dummy_model),
            X_y_func=X_y_func,
            metric=copy.deepcopy(metric)
        )


def check_predict_proba_one_binary(classifier):

    for x, y in make_random_X_y(classifier, n_observations=20, n_features=4):
        y_pred = classifier.predict_proba_one(x)
        classifier = classifier.fit_one(x, y)
        assert len(y_pred) == 2
        assert True in y_pred
        assert False in y_pred


def yield_all_checks(model):

    from .. import base

    yield check_fit_one

    if isinstance(model, base.Classifier):
        yield check_predict_proba_one

    # MultiClassifiers are also BinaryClassifiers so binary will apply to both
    if isinstance(model, base.BinaryClassifier):
        yield check_better_than_dummy_binary

        # Some tests work for BinaryClassifiers but not for MultiClassifiers
        if not isinstance(model, base.MultiClassifier):
            yield check_predict_proba_one_binary

    if isinstance(model, base.MultiClassifier):
        yield check_better_than_dummy_multi

    if isinstance(model, base.Regressor):
        yield check_better_than_dummy_regression


def check_estimator(model):

    for check in yield_all_checks(model):
        check(copy.deepcopy(model))



import abc
import math

import numpy as np

from .. import utils


__all__ = [
    'AbsoluteLoss',
    'BinaryClassificationLoss',
    'CauchyLoss',
    'CrossEntropy',
    'HingeLoss',
    'LogLoss',
    'MultiClassificationLoss',
    'RegressionLoss',
    'SquaredLoss'
]


def clip_proba(p):
    return utils.clip(p, minimum=1e-15, maximum=1 - 1e-15)


class Loss(abc.ABC):

    @abc.abstractmethod
    def __call__(self, y_true, y_pred) -> float:
        """Returns the loss."""

    @abc.abstractmethod
    def gradient(self, y_true, y_pred) -> float:
        """Returns the gradient with respect to ``y_pred``."""


class BinaryClassificationLoss(Loss):
    """A loss appropriate binary classification tasks."""


class MultiClassificationLoss(Loss):
    """A loss appropriate for multi-class classification tasks."""


class RegressionLoss(Loss):
    """A loss appropriate for regression tasks."""


class AbsoluteLoss(RegressionLoss):
    """Computes the absolute loss, also known as the mean absolute error or L1 loss.

    Mathematically, it is defined as

    .. math:: L = |p_i - y_i|

    It's gradient w.r.t. to $p_i$ is

    .. math:: \\frac{\\partial L}{\\partial p_i} = sgn(p_i - y_i)

    Example:

        ::

            >>> from creme import optim

            >>> loss = optim.AbsoluteLoss()
            >>> loss(-42, 42)
            84
            >>> loss.gradient(1, 2)
            1
            >>> loss.gradient(2, 1)
            -1

    """

    def __call__(self, y_true, y_pred):
        return abs(y_pred - y_true)

    def gradient(self, y_true, y_pred):
        return np.sign(y_pred - y_true)


class CauchyLoss(RegressionLoss):
    """Cauchy loss function.

    References:

        1. `Effect of MAE <https://www.kaggle.com/c/allstate-claims-severity/discussion/24520#140163>`_
        2. `Paris Madness <https://www.kaggle.com/raddar/paris-madness>`_

    """

    def __init__(self, C=80):
        self.C = C

    def __call__(self, y_true, y_pred):
        return abs(y_pred - y_true)

    def gradient(self, y_true, y_pred):
        diff = y_pred - y_true
        return diff / ((diff / self.C) ** 2 + 1)


class SquaredLoss(RegressionLoss):
    """Computes the squared loss, also known as the L2 loss.

    Mathematically, it is defined as

    .. math:: L = \\frac{(p_i - y_i)^2}{2}

    It's gradient w.r.t. to $p_i$ is

    .. math:: \\frac{\\partial L}{\\partial p_i} = p_i - y_i

    Example:

        ::

            >>> from creme import optim

            >>> loss = optim.SquaredLoss()
            >>> loss(-4, 5)
            40.5
            >>> loss.gradient(1, 4)
            3
            >>> loss.gradient(4, 1)
            -3

    """

    def __call__(self, y_true, y_pred):
        return .5 * (y_pred - y_true) * (y_pred - y_true)

    def gradient(self, y_true, y_pred):
        return y_pred - y_true


class LogLoss(BinaryClassificationLoss):
    """Computes the logarithmic loss.

    Mathematically, it is defined as

    .. math:: L = -y_i log(p_i) + (1-y_i) log(1-p_i)

    It's gradient w.r.t. to $p_i$ is

    .. math:: \\frac{\\partial L}{\\partial p_i} = sign(p_i - y_i)

    Example:

        ::

            >>> from creme import optim

            >>> loss = optim.LogLoss()
            >>> loss(1, 0.5)
            0.693147...
            >>> loss(1, 0)
            34.53877...
            >>> loss.gradient(1, 0.2)
            -0.8
            >>> loss.gradient(0, 0.2)
            0.2

    """

    def __call__(self, y_true, y_pred):
        y_pred = clip_proba(y_pred)
        if y_true:
            return -math.log(y_pred)
        return -math.log(1 - y_pred)

    def gradient(self, y_true, y_pred):
        return clip_proba(y_pred) - y_true


class HingeLoss(BinaryClassificationLoss):
    """Computes the hinge loss.

    Mathematically, it is defined as

    .. math:: L = max(0, 1 - p_i * y_i)


    It's gradient w.r.t. to $p_i$ is

    .. math::
        \\frac{\\partial L}{\\partial y_i} = \\left\{
        \\begin{array}{ll}
            \\ 0  &   p_iy_i \geqslant 1  \\\\
            \\ - y_i & p_iy_i < 1
        \\end{array}
        \\right.

    Example:

        ::

            >>> from creme import optim
            >>> import numpy as np
            >>> from sklearn import svm
            >>> from sklearn.metrics import hinge_loss

            >>> X = [[0], [1]]
            >>> y = [-1, 1]
            >>> lin_svm = svm.LinearSVC(random_state=0).fit(X, y)

            >>> y_true = [0, 1, 1]
            >>> y_pred = lin_svm.decision_function([[-2], [3], [0.5]])

            >>> hinge_loss([0, 1, 1], y_pred)
            0.303036...

            >>> loss = optim.HingeLoss()
            >>> np.mean([loss(y_t, y_p) for y_t, y_p in zip(y_true, y_pred)])
            0.303036...

    """

    def __call__(self, y_true, y_pred):
        # Our convention is to use 0s instead of -1s for negatives, but the Hinge loss uses -1s as
        # a convention
        y_true = y_true or -1
        return max(0, 1 - y_true * y_pred)

    def gradient(self, y_true, y_pred):
        """Returns the gradient with respect to ``y_pred``.

        References:

            1. `Wolfram Alpha <https://www.wolframalpha.com/input/?i=derivative+max(0,+1+-+p+*+y)+wrt+p>`_

        """
        y_true = y_true or -1
        if y_true * y_pred < 1:
            return -y_pred
        return 0


class EpsilonInsensitiveHingeLoss(RegressionLoss):

    def __init__(self, eps=0.1):
        self.eps = eps

    def __call__(self, y_true, y_pred):
        # Our convention is to use 0s instead of -1s for negatives, but the Hinge loss uses -1s as
        # a convention
        y_true = y_true or -1
        return max(0, abs(y_pred - y_true) - self.eps)

    def gradient(self, y_true, y_pred):
        """Returns the gradient with respect to ``y_pred``.

        References:

            1. `Wolfram Alpha <https://www.wolframalpha.com/input/?i=derivative+max(0,+abs(p+-+y)+-+eps)+wrt+p>`_

        """
        y_true = y_true or -1
        if y_pred > y_true + self.eps:
            return 1
        elif y_pred + self.eps < y_true:
            return -1
        return 0


class CrossEntropy(MultiClassificationLoss):
    """Cross entropy is a generalization of logistic loss to multiple classes.

    Example:

        ::

            >>> from creme import optim

            >>> y_true = [0, 1, 2, 2]
            >>> y_pred = [
            ...     {0: 0.29450637, 1: 0.34216758, 2: 0.36332605},
            ...     {0: 0.21290077, 1: 0.32728332, 2: 0.45981591},
            ...     {0: 0.42860913, 1: 0.33380113, 2: 0.23758974},
            ...     {0: 0.44941979, 1: 0.32962558, 2: 0.22095463}
            ... ]

            >>> loss = optim.CrossEntropy()

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(loss(y_t, y_p))
            1.222454...
            1.116929...
            1.437209...
            1.509797...

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(loss.gradient(y_t, y_p))
            {0: -0.70549363, 1: 0.34216758, 2: 0.36332605}
            {0: 0.21290077, 1: -0.67271668, 2: 0.45981591}
            {0: 0.42860913, 1: 0.33380113, 2: -0.76241026}
            {0: 0.44941979, 1: 0.32962558, 2: -0.77904537}

    References:

        1. `What is Softmax regression and how is it related to Logistic regression? <https://github.com/rasbt/python-machine-learning-book/blob/master/faq/softmax_regression.md>`_

    """

    def __call__(self, y_true, y_pred):
        return -sum(
            (y_true == label) * math.log(clip_proba(proba))
            for label, proba in y_pred.items()
        )

    def gradient(self, y_true, y_pred):
        return {
            label: clip_proba(y_pred.get(label, 0.)) - (y_true == label)
            for label in {*y_pred.keys(), y_true}
        }



import collections

from .. import base
from .. import optim
from .. import stats
from .. import utils


__all__ = ['LinearRegression']


class LinearRegression(base.Regressor):
    """Linear regression.

    A linear regression is simply a dot product between some features and some weights. The weights
    are found by using an online optimizer. The intercept is computed independently by maintaining
    a running mean and adding it to each prediction. Although this isn't the textbook way of doing
    online linear regression, it works just as well if not better.

    Parameters:
        optimizer (optim.Optimizer): The sequential optimizer used to find the best weights.
            Defaults to `optim.VanillaSGD`.
        loss (optim.RegressionLoss): The loss function to minimize. Defaults to
            `optim.SquaredLoss`.
        l2 (float): Amount of L2 regularization used to push weights towards 0.
        intercept (stats.Univariate): The univariate statistic used to compute the intercept
            online. Defaults to `stats.Mean`.

    Attributes:
        weights (collections.defaultdict)
            Coefficients assigned to each feature.

    Example:

        ::

            >>> from creme import compose
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     load_dataset=datasets.load_boston,
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> model = compose.Pipeline([
            ...     ('scale', preprocessing.StandardScaler()),
            ...     ('lin_reg', linear_model.LinearRegression(
            ...         loss=optim.CauchyLoss()
            ...     ))
            ... ])
            >>> metric = metrics.MAE()

            >>> model_selection.online_score(X_y, model, metric)
            MAE: 3.667428

            >>> model['lin_reg'].intercept.get()
            22.532806...

    Note:
        Using a feature scaler such as `preprocessing.StandardScaler` upstream helps the optimizer
        to converge.

    """

    def __init__(self, optimizer=None, loss=None, l2=0., intercept=None):
        self.optimizer = optim.VanillaSGD(0.01) if optimizer is None else optimizer
        self.loss = optim.SquaredLoss() if loss is None else loss
        self.l2 = l2
        self.intercept = stats.Mean() if intercept is None else intercept
        self.weights = collections.defaultdict(float)

    def fit_one(self, x, y):

        # Some optimizers need to do something before a prediction is made
        self.weights = self.optimizer.update_before_pred(w=self.weights)

        # Make a prediction for the given features
        y_pred = self.predict_one(x)

        # Compute the gradient w.r.t. each feature
        loss_gradient = self.loss.gradient(y_true=y, y_pred=y_pred)
        gradient = {
            i: xi * loss_gradient + self.l2 * self.weights.get(i, 0)
            for i, xi in x.items()
        }

        # Update the weights by using the gradient
        self.weights = self.optimizer.update_after_pred(g=gradient, w=self.weights)

        # Update the intercept
        if self.intercept:
            self.intercept.update(y)

        return self

    def predict_one(self, x):
        y = utils.dot(x, self.weights)
        if self.intercept:
            y += self.intercept.get()
        return y



import abc

from .. import base
from .. import types


__all__ = [
    'BinaryClassificationMetric',
    'MultiClassificationMetric',
    'RegressionMetric'
]


class BaseMetric(abc.ABC):

    @abc.abstractmethod
    def get(self) -> float:
        """Returns the current value of the metric."""

    @abc.abstractmethod
    def works_with(self, model) -> bool:
        """Tells if a metric can work with a given model or not."""

    @property
    @abc.abstractmethod
    def bigger_is_better(self) -> bool:
        """Indicates if a high value is better than a low one or not."""

    def __str__(self):
        """Returns the class name along with the current value of the metric."""
        return f'{self.__class__.__name__}: {self.get():.6f}'.rstrip('0')

    def __repr__(self):
        return str(self)


class ClassificationMetric(BaseMetric):

    @property
    @abc.abstractmethod
    def requires_labels(self):
        """Helps to indicate if labels are required instead of probabilities."""


class BinaryClassificationMetric(ClassificationMetric):

    @abc.abstractmethod
    def update(self, y_true: bool, y_pred: float):
        """Updates the metric."""

    def works_with(self, model):
        return isinstance(model, base.BinaryClassifier)


class MultiClassificationMetric(BinaryClassificationMetric):

    @abc.abstractmethod
    def update(self, y_true: types.Label, y_pred: dict):
        """Updates the metric."""

    def works_with(self, model):
        return isinstance(model, (base.BinaryClassifier, base.MultiClassifier))


class RegressionMetric(BaseMetric):

    @abc.abstractmethod
    def update(self, y_true: float, y_pred: float):
        """Updates the metric."""

    def works_with(self, model):
        return isinstance(model, base.Regressor)



import abc


__all__ = [
    'ConstantLR',
    'InverseScalingLR',
    'OptimalLR'
]


class LRScheduler(abc.ABC):

    @abc.abstractmethod
    def get(self, t: int) -> float:
        pass


class ConstantLR(LRScheduler):
    """Always uses the same learning rate."""

    def __init__(self, learning_rate):
        self.learning_rate = learning_rate

    def get(self, t):
        return self.learning_rate


class InverseScalingLR(LRScheduler):
    """Reduces the learning rate using a power schedule.

    Assuming an iteration counter $t$ starting from 0, the learning rate will be:

    .. math:: \\frac{1}{(t+1)^p}

    where $p$ is a user-defined parameter.

    """

    def __init__(self, learning_rate, power=0.5):
        self.learning_rate = learning_rate
        self.power = power

    def get(self, t):
        return self.learning_rate / (t + 1) ** self.power


class OptimalLR(LRScheduler):
    """Optimal learning schedule as proposed by Léon Bottou.

    References:

        1. `Stochastic Gradient Descent <https://leon.bottou.org/projects/sgd>`_

    """

    def __init__(self, t0=1000, alpha=1e-4):
        self.t0 = t0
        self.alpha = alpha

    def get(self, t):
        return 1. / (self.alpha * (t + self.t0))



from . import base


class Mean(base.Univariate):
    """Running mean.

    Attributes:
        mean (float): The running mean.
        n (int): The current number of observations.

    Example:

        ::

            >>> from creme import stats

            >>> X = [-5, -3, -1, 1, 3, 5]
            >>> mean = stats.Mean()
            >>> for x in X:
            ...     print(mean.update(x).get())
            -5.0
            -4.0
            -3.0
            -2.0
            -1.0
            0.0

    """

    def __init__(self, n=0, mean=0):
        self.n = n
        self.mean = mean

    @property
    def name(self):
        return 'mean'

    def update(self, x):
        self.n += 1
        self.mean += (x - self.mean) / self.n if self.n > 0 else 0
        return self

    def get(self):
        return self.mean

    def __add__(self, other):
        return Mean(n=1, mean=self.mean + other.mean)

    def __mul__(self, other):
        return Mean(n=self.n, mean=self.mean * other)


class BayesianMean(base.Univariate):
    """Estimates a mean using outside information.

    References:

        1. `Additive smoothing <https://www.wikiwand.com/en/Additive_smoothing>`_
        2. `Bayesian average <https://www.wikiwand.com/en/Bayesian_average>`_
        3. `Practical example of Bayes estimators <https://www.wikiwand.com/en/Bayes_estimator#/Practical_example_of_Bayes_estimators>`_

    """

    def __init__(self, prior, prior_weight):
        self.prior = prior
        self.prior_weight = prior_weight
        self.mean = Mean()

    @property
    def name(self):
        return 'bayes_mean'

    def update(self, x):
        self.mean.update(x)
        return self

    def get(self):

        # Uses the notation from https://www.wikiwand.com/en/Bayes_estimator#/Practical_example_of_Bayes_estimators
        R = self.mean.get()
        v = self.mean.n
        m = self.prior_weight
        C = self.prior

        return (R * v + C * m) / (v + m)



import collections


__all__ = ['ConfusionMatrix']


class ConfusionMatrix(collections.defaultdict):
    """Confusion matrix.

    This class is different from the rest of the classes from the :any:`metrics` module in that it
    doesn't have a ``get`` method. The only use of this class is to be able to elegantly printable.

    Attributes:
        classes (set): The entire set of seen classes, whether they are part of the predictions or
            the true values.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = ['cat', 'ant', 'cat', 'cat', 'ant', 'bird']
            >>> y_pred = ['ant', 'ant', 'cat', 'cat', 'ant', 'cat']

            >>> cm = metrics.ConfusionMatrix()

            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     cm = cm.update(y_t, y_p)

            >>> cm
                     ant  bird   cat
               ant     2     0     0
              bird     0     0     1
               cat     1     0     2

    """

    def __init__(self):
        super().__init__(collections.Counter)
        self.classes = set()

    def update(self, y_true, y_pred):
        self[y_true].update([y_pred])
        self.classes.update({y_true, y_pred})
        return self

    def __str__(self):

        # The classes are sorted alphabetically for reproducibility reasons
        classes = sorted(self.classes)

        # Determine the required width of each column in the table
        largest_label_len = max(map(len, classes))
        largest_number_len = len(str(max(max(counter.values()) for counter in self.values())))
        width = max(largest_label_len, largest_number_len) + 2

        # Make a template to print out rows one by one
        row_format = '{:>{width}}' * (len(classes) + 1)

        # Write down the header
        table = row_format.format('', *classes, width=width) + '\n'

        # Write down the true labels row by row
        table += '\n'.join((
            row_format.format(y_true, *[self[y_true][y_pred] for y_pred in classes], width=width)
            for y_true in classes
        ))

        return table

    def __repr__(self):
        return str(self)



from .. import utils

from . import base


class RollingMin(base.Univariate, utils.SortedWindow):
    """Running min over a window.

    Parameters:
        window_size (int): Size of the rolling window.

    Example:

        ::

            >>> from creme import stats

            >>> X = [1, -4, 3, -2, 2, 1]
            >>> rolling_min = stats.RollingMin(2)
            >>> for x in X:
            ...     print(rolling_min.update(x).get())
            1
            -4
            -4
            -2
            -2
            1

    """

    @property
    def name(self):
        return f'rolling_{self.window_size}_min'

    def update(self, x):
        self.append(x)
        return self

    def get(self):
        return self[0]



import collections
import copy
import statistics

from sklearn import utils

from .. import base


__all__ = ['BaggingClassifier', 'BaggingRegressor']


class BaggingClassifier(base.Classifier):
    """Bagging for classification.

    For each incoming observation, each model's ``fit_one`` method is called ``k`` times where
    ``k`` is sampled from a Poisson distribution of parameter 1. ``k`` thus has a 36% chance of
    being equal to 0, a 36% chance of being equal to 1, an 18% chance of being equal to 2, a 6%
    chance of being equal to 3, a 1% chance of being equal to 4, etc. You can do
    ``scipy.stats.poisson(1).pmf(k)`` to obtain more detailed values.

    Parameters:
        classifier (BinaryClassifier or MultiClassifier): The classifier to bag.

    Example:

        In the following example three logistic regressions are bagged together. The performance is
        slightly better than when using a single logistic regression.

        ::

            >>> from creme import compose
            >>> from creme import ensemble
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     load_dataset=datasets.load_breast_cancer,
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> optimiser = optim.VanillaSGD()
            >>> model = compose.Pipeline([
            ...     ('scale', preprocessing.StandardScaler()),
            ...     ('learn', linear_model.LogisticRegression(optimiser))
            ... ])
            >>> model = ensemble.BaggingClassifier(
            ...     classifier=model,
            ...     n_classifiers=3,
            ...     random_state=42
            ... )
            >>> metric = metrics.F1Score()

            >>> model_selection.online_score(X_y, model, metric)
            F1Score: 0.970464

    References:

        1. `Online Bagging and Boosting <https://ti.arc.nasa.gov/m/profile/oza/files/ozru01a.pdf>`_

    """

    def __init__(self, classifier=None, n_classifiers=10, random_state=None):
        self.classifier = classifier
        self.classifiers = [copy.deepcopy(classifier) for _ in range(n_classifiers)]
        self.rng = utils.check_random_state(random_state)

    @property
    def __class__(self):
        return self.classifier.__class__

    def __str__(self):
        return f'BaggingClassifier({str(self.classifier)})'

    def fit_one(self, x, y):

        for classifier in self.classifiers:
            for _ in range(self.rng.poisson(1)):
                classifier.fit_one(x, y)

        return self

    def predict_proba_one(self, x):
        """Averages the predictions of each classifier."""

        y_pred = collections.defaultdict(float)

        # Sum the predictions
        for classifier in self.classifiers:
            for label, proba in classifier.predict_proba_one(x).items():
                y_pred[label] += proba

        # Divide by the number of predictions
        for label in y_pred:
            y_pred[label] /= len(self.classifiers)

        return dict(y_pred)


class BaggingRegressor(base.Regressor):
    """Bagging for regression.

    For each incoming observation, each model's `fit_one` method is called ``k`` times where ``k``
    is sampled from a Poisson distribution of parameter 1. ``k`` thus has a 36% chance of being
    equal to 0, a 36% chance of being equal to 1, an 18% chance of being equal to 2, a 6% chance of
    being equal to 3, a 1% chance of being equal to 4, etc. You can do
    ``scipy.stats.poisson(1).pmf(k)`` for more detailed values.

    Parameters:
        base_regressor (Regressor): The regressor to bag.

    Example:

        In the following example three logistic regressions are bagged together. The performance is
        slightly better than when using a single logistic regression.

        ::

            >>> from creme import compose
            >>> from creme import ensemble
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     load_dataset=datasets.load_boston,
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> model = compose.Pipeline([
            ...     ('scale', preprocessing.StandardScaler()),
            ...     ('learn', linear_model.LinearRegression())
            ... ])
            >>> metric = metrics.MSE()

    References:

        1. `Online Bagging and Boosting <https://ti.arc.nasa.gov/m/profile/oza/files/ozru01a.pdf>`_

    """

    def __init__(self, regressor=None, n_regressors=10, random_state=42):
        self.regressor = regressor
        self.regressors = [copy.deepcopy(regressor) for _ in range(n_regressors)]
        self.rng = utils.check_random_state(random_state)

    def __str__(self):
        return f'BaggingRegressor({str(self.regressor)})'

    def fit_one(self, x, y):

        for regressor in self.regressors:
            for _ in range(self.rng.poisson(1)):
                regressor.fit_one(x, y)

        return self

    def predict_one(self, x):
        """Averages the predictions of each regressor."""
        return statistics.mean((regressor.predict_one(x) for regressor in self.regressors))



import math

from . import base


class Min(base.Univariate):
    """Running min.

    Attributes:
        min (float): The running min.

    """

    def __init__(self):
        self.min = math.inf

    @property
    def name(self):
        return 'min'

    def update(self, x):
        if x < self.min:
            self.min = x
        return self

    def get(self):
        return self.min



import collections

from .. import base
from .. import optim
from .. import utils


__all__ = ['LogisticRegression']


class LogisticRegression(base.BinaryClassifier):
    """Logistic regression for binary classification.

    Parameters:
        optimizer (optim.Optimizer): The sequential optimizer used to find the best weights.
        loss (optim.BinaryClassificationLoss): The loss function to optimize for.
        l2 (float): Amount of L2 regularization used to push weights towards 0.

    Attributes:
        weights (collections.defaultdict)

    Example:

        ::

            >>> from creme import compose
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     load_dataset=datasets.load_breast_cancer,
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> model = compose.Pipeline([
            ...     ('scale', preprocessing.StandardScaler()),
            ...     ('learn', linear_model.LogisticRegression())
            ... ])
            >>> metric = metrics.F1Score()

            >>> model_selection.online_score(X_y, model, metric)
            F1Score: 0.964739

    """

    def __init__(self, optimizer=None, loss=None, l2=0):
        self.optimizer = optim.VanillaSGD(0.01) if optimizer is None else optimizer
        self.loss = optim.LogLoss() if loss is None else loss
        self.l2 = l2
        self.weights = collections.defaultdict(float)

    def fit_one(self, x, y):

        # Some optimizers need to do something before a prediction is made
        self.weights = self.optimizer.update_before_pred(w=self.weights)

        # Make a prediction for the given features
        y_pred = self.predict_proba_one(x)[True]

        # Compute the gradient w.r.t. each feature
        loss_gradient = self.loss.gradient(y_true=y, y_pred=y_pred)
        gradient = {
            i: xi * loss_gradient + self.l2 * self.weights.get(i, 0)
            for i, xi in x.items()
        }

        # Update the weights by using the gradient
        self.weights = self.optimizer.update_after_pred(g=gradient, w=self.weights)

        return self

    def predict_proba_one(self, x):
        y_pred = utils.sigmoid(utils.dot(x, self.weights))
        return {False: 1. - y_pred, True: y_pred}



"""
Utilities for streaming data from various sources.
"""
import csv
import datetime as dt
import itertools

from sklearn import utils


__all__ = [
    'iter_csv',
    'iter_numpy',
    'iter_pandas',
    'iter_sklearn_dataset'
]


def iter_numpy(X, y=None, feature_names=None, shuffle=False, random_state=None):
    """Yields rows from an array of features and an array of targets.

    Parameters:
        X (array-like of shape (n_samples, n_features))
        y (array-like of shape (n_samples,))
        feature_names (list of length n_features)
        shuffle (bool): Whether to shuffle the inputs or not.
        random_state (int, RandomState instance or None, default=None): If int, ``random_state`` is
            the seed used by the random number generator; if ``RandomState`` instance,
            ``random_state`` is the random number generator; if ``None``, the random number
            generator is the ``RandomState`` instance used by ``np.random``.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    """
    feature_names = list(range(len(X[0]))) if feature_names is None else feature_names
    rng = utils.check_random_state(random_state)

    if shuffle:
        order = rng.permutation(len(X))
        X, y = X[order], y if y is None else y[order]

    for x, yi in itertools.zip_longest(X, y if hasattr(y, '__iter__') else []):
        yield {i: xi for i, xi in zip(feature_names, x)}, yi


def iter_sklearn_dataset(load_dataset, **kwargs):
    """Yields rows from one of the datasets provided by scikit-learn.

    Parameters:
        load_dataset (callable): The method used to load the dataset, for example
        `sklearn.datasets.load_boston`.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    """
    dataset = load_dataset()
    kwargs['X'] = dataset.data
    kwargs['y'] = dataset.target
    kwargs['feature_names'] = dataset.feature_names

    for x, yi in iter_numpy(**kwargs):
        yield x, yi


def iter_pandas(X, y=None, **kwargs):
    """Yields rows from a ``pandas.DataFrame``.

    Parameters:
        X (pandas.DataFrame)
        y (array-like of shape (n_samples,))

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    """
    kwargs['feature_names'] = X.columns
    for x, yi in iter_numpy(X, y, **kwargs):
        yield x, yi


def iter_csv(filepath_or_buffer, target_name, types=None, parse_dates=None):
    """Yields rows from a CSV file.

    Parameters:
        filepath_or_buffer: Either a string indicating the location of a CSV file, or a buffer
            object that has a ``read`` method.
        types (dict): The type of each feature.
        parse_dates (dict): A `dict` mapping feature names to a format passed to the
            `datetime.datetime.strptime` method.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    Example:

        ::

            >>> import io
            >>> from creme import stream

            >>> data = io.StringIO('''name,day,viewers
            ... Breaking Bad,2018-03-14,1337
            ... The Sopranos,2018-03-14,42
            ... Breaking Bad,2018-03-15,7331
            ... ''')

            >>> params = dict(
            ...     target_name='viewers',
            ...     types={'viewers': int},
            ...     parse_dates={'day': '%Y-%m-%d'}
            ... )

            >>> for x, y in stream.iter_csv(data, **params):
            ...     print(x, y)
            OrderedDict([('name', 'Breaking Bad'), ('day', datetime.datetime(2018, 3, 14, 0, 0))]) 1337
            OrderedDict([('name', 'The Sopranos'), ('day', datetime.datetime(2018, 3, 14, 0, 0))]) 42
            OrderedDict([('name', 'Breaking Bad'), ('day', datetime.datetime(2018, 3, 15, 0, 0))]) 7331

    """

    file = filepath_or_buffer

    if not hasattr(file, 'read'):
        file = open(file)

    for x in csv.DictReader(file):

        # Cast the values to the given types
        if types is not None:
            for i, t in types.items():
                x[i] = t(x[i])

        # Parse the dates
        if parse_dates is not None:
            for i, fmt in parse_dates.items():
                x[i] = dt.datetime.strptime(x[i], fmt)

        # Separate the target from the features
        y = x.pop(target_name)

        yield x, y



import collections
import statistics

from .. import stats

from . import base


__all__ = ['Precision', 'MacroPrecision', 'MicroPrecision']


class Precision(stats.Mean, base.BinaryClassificationMetric):
    """Binary precision score.

    Example:

        ::

            >>> from creme import metrics
            >>> from sklearn.metrics import precision_score

            >>> y_true = [True, False, True, True, True]
            >>> y_pred = [True, True, False, True, True]

            >>> metric = metrics.Precision()
            >>> for i, (y_t, y_p) in enumerate(zip(y_true, y_pred)):
            ...     metric = metric.update(y_t, y_p)
            ...     assert metric.get() == precision_score(y_true[:i+1], y_pred[:i+1])

            >>> metric
            Precision: 0.75

    """

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True

    def update(self, y_true, y_pred):
        if y_pred:
            return super().update(y_true == y_pred)
        return self


class MacroPrecision(base.MultiClassificationMetric):
    """Macro-average precision score.

    Example:

        ::

            >>> from creme import metrics
            >>> from sklearn.metrics import precision_score

            >>> y_true = [0, 1, 2, 2, 2]
            >>> y_pred = [0, 0, 2, 2, 1]

            >>> metric = metrics.MacroPrecision()
            >>> for i, (y_t, y_p) in enumerate(zip(y_true, y_pred)):
            ...     metric = metric.update(y_t, y_p)
            ...     print(metric.get(), precision_score(y_true[:i+1], y_pred[:i+1], average='macro'))
            1.0 1.0
            0.25 0.25
            0.5 0.5
            0.5 0.5
            0.5 0.5

            >>> metric
            MacroPrecision: 0.5

    """

    def __init__(self):
        self.precisions = collections.defaultdict(Precision)
        self.classes = set()

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True

    def update(self, y_true, y_pred):
        self.precisions[y_pred].update(y_true == y_pred, True)
        self.classes.update({y_true, y_pred})
        return self

    def get(self):
        return statistics.mean((
            0 if c not in self.precisions else self.precisions[c].get()
            for c in self.classes
        ))


class MicroPrecision(stats.Mean, base.MultiClassificationMetric):
    """Micro-average precision score.

    The micro-average precision score is exactly equivalent to the micro-average recall as well as
    the micro-average F1 score.

    Example:

        ::

            >>> from creme import metrics
            >>> from sklearn.metrics import precision_score

            >>> y_true = [0, 1, 2, 2, 2]
            >>> y_pred = [0, 0, 2, 2, 1]

            >>> metric = metrics.MicroPrecision()
            >>> for i, (y_t, y_p) in enumerate(zip(y_true, y_pred)):
            ...     metric = metric.update(y_t, y_p)
            ...     print(metric.get(), precision_score(y_true[:i+1], y_pred[:i+1], average='micro'))
            1.0 1.0
            0.5 0.5
            0.666666... 0.666666...
            0.75 0.75
            0.6 0.6

            >>> metric
            MicroPrecision: 0.6

    References:

        1. `Why are precision, recall and F1 score equal when using micro averaging in a multi-class problem? <https://simonhessner.de/why-are-precision-recall-and-f1-score-equal-when-using-micro-averaging-in-a-multi-class-problem/>`_

    """

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True

    def update(self, y_true, y_pred):
        super().update(y_true == y_pred)
        return self



import os
import shutil
import urllib
import zipfile

from .. import stream


def get_data_home(data_home=None):
    """Return the path of the creme data directory."""
    if data_home is None:
        data_home = os.environ.get('CREME_DATA', os.path.join('~', 'creme_data'))
        data_home = os.path.expanduser(data_home)
        if not os.path.exists(data_home):
            os.makedirs(data_home)
    return data_home


def fetch_csv_dataset(data_home, url, name, **iter_csv_params):

    data_home = get_data_home(data_home=data_home)

    # If the CSV file exists then iterate over it
    csv_path = os.path.join(data_home, f'{name}.csv')
    if os.path.exists(csv_path):
        return stream.iter_csv(csv_path, **iter_csv_params)

    # If the ZIP file exists then unzip it
    zip_path = os.path.join(data_home, f'{name}.zip')
    if os.path.exists(zip_path):
        print('Unzipping data...')

        # Unzip the ZIP file
        with zipfile.ZipFile(zip_path, 'r') as zf:
            zf.extractall(data_home)

        # Delete the ZIP file now that the CSV file is available
        os.remove(zip_path)

        return fetch_csv_dataset(data_home, url, name, **iter_csv_params)

    # Download the ZIP file
    print('Downloading data...')
    with urllib.request.urlopen(url) as r, open(zip_path, 'wb') as f:
        shutil.copyfileobj(r, f)

    return fetch_csv_dataset(data_home, url, name, **iter_csv_params)



import numpy as np

from .. import utils


class SDFT(utils.Window):
    """Sliding Discrete Fourier Transform (SDFT).

    Initially, the coefficients are all equal to 0, up until enough values have been seen. A call
    to `numpy.fft.fft` is triggered once ``window_size`` values have been seen. Subsequent values
    will update the coefficients online. This is much faster than recomputing an FFT from scratch
    for every new value.

    Parameters:
        window_size (int): The size of the window.

    Attributes:
        fft (numpy array of complex numbers): The Fourier components.

    Example:

        ::

            >>> X = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

            >>> window_size = 5
            >>> sdft = SDFT(window_size)

            >>> for i, x in enumerate(X):
            ...     sdft = sdft.update(x)
            ...
            ...     if i + 1 >= window_size:
            ...         assert np.allclose(
            ...             sdft.fft,
            ...             np.fft.fft(X[i+1 - window_size:i+1])
            ...         )

    """

    def __init__(self, window_size):
        super().__init__(window_size=window_size)
        self.fft = np.zeros(window_size)

    def update(self, x):

        # Simply append the new value if the window isn't full yet
        if len(self) < self.window_size - 1:
            self.append(x)

        # Compute an initial FFT the first time the window is full
        elif len(self) == self.window_size - 1:
            self.append(x)
            self.fft = np.fft.fft(self).tolist()

        # Update the coefficients for subsequent values
        else:
            diff = x - self[0]
            for i in range(self.window_size):
                self.fft[i] = (self.fft[i] + diff) * np.exp(2j * np.pi * i / self.window_size)
            self.append(x)

        return self



import functools
import math
import operator

import numpy as np


__all__ = [
    'chain_dot',
    'clip',
    'dot',
    'norm',
    'prod',
    'sigmoid',
    'softmax'
]


def softmax(y_pred):
    """Normalizes a dictionary of predicted probabilities, in-place."""
    exp = {c: math.exp(p) for c, p in y_pred.items()}
    total = sum(exp.values())
    return {c: exp[c] / total for c in y_pred}


def prod(iterable):
    return functools.reduce(operator.mul, iterable, 1)


def dot(x: dict, y: dict):
    """Returns the dot product of two vectors represented as dicts.

    Example:

        >>> x = {'x0': 1, 'x1': 2}
        >>> y = {'x1': 21, 'x2': 3}
        >>> dot(x, y)
        42

    """
    return sum(xi * y.get(i, 0) for i, xi in min(x, y, key=len).items())


def chain_dot(*xs):
    """Returns the dot product of multiple vectors represented as dicts.

    Example:

        >>> x = {'x0': 1, 'x1': 2, 'x2': 1}
        >>> y = {'x1': 21, 'x2': 3}
        >>> z = {'x1': 2, 'x2': 1 / 3}
        >>> chain_dot(x, y, z)
        85.0

    """
    keys = min(xs, key=len)
    return sum(prod(x.get(i, 0) for x in xs) for i in keys)


def sigmoid(x: float):
    if x < -30:
        return 0
    if x > 30:
        return 1
    return 1 / (1 + math.exp(-x))


def clip(x: float, minimum=0., maximum=1.):
    return max(min(x, maximum), minimum)


def norm(x, order=None):
    return np.linalg.norm(list(x.values()), ord=order)



from .. import stats

from . import base


__all__ = ['MAE']


class MAE(stats.Mean, base.RegressionMetric):
    """Mean absolute error.

    Example:

        ::

            >>> import math
            >>> from creme import metrics
            >>> from sklearn.metrics import mean_absolute_error

            >>> y_true = [3, -0.5, 2, 7]
            >>> y_pred = [2.5, 0.0, 2, 8]

            >>> metric = metrics.MAE()
            >>> for i, (y_t, y_p) in enumerate(zip(y_true, y_pred)):
            ...     metric = metric.update(y_t, y_p)
            ...     assert math.isclose(
            ...         metric.get(),
            ...         mean_absolute_error(y_true[:i+1], y_pred[:i+1])
            ...     )

            >>> metric
            MAE: 0.5

    """

    @property
    def bigger_is_better(self):
        return False

    def update(self, y_true, y_pred):
        return super().update(abs(y_true - y_pred))



import copy
import math

from .. import stats

from . import base


class Normal(base.ContinuousDistribution):
    """Normal distribution with parameters mu and sigma.

    Example:

        ::

            >>> d1 = Normal().update(5)
            >>> d1.mode()
            5.0

            >>> d2 = Normal().update(6).update(7)
            >>> d2.mode()
            6.5

            >>> d1 += d2
            >>> d1.mode()
            11.5

            >>> d2.pdf(6.5)
            0.564189...

    """

    def __init__(self):
        self.variance = stats.Variance()

    @property
    def mu(self):
        return self.variance.mean.get()

    @property
    def sigma(self):
        return self.variance.get() ** 0.5

    def update(self, x):
        self.variance.update(x)
        return self

    def mode(self):
        return self.mu

    def pdf(self, x):
        variance = self.variance.get()

        if variance == 0:
            return 0

        return math.exp((x - self.mu) ** 2 / (-2 * variance)) / math.sqrt(math.tau * variance)

    def __iadd__(self, other):
        if isinstance(other, Normal):
            other = other.variance.mean
        self.variance.mean += other
        return self

    def __imul__(self, constant):
        self.variance.mean *= constant
        return self

    def __copy__(self):
        """Custom copying to make the ``copy.copy`` method faster."""
        clone = Normal()
        clone.variance = copy.copy(self.variance)
        return clone

    def __deepcopy__(self, memo):
        return self.__copy__()



from .. import utils

from . import base


class RollingQuantile(base.Univariate, utils.SortedWindow):
    """Running quantile over a window.

    Parameters:
        window_size (int): Size of the window.
        quantile (float): Desired quantile, must be between 0 and 1.

    Attributes:
        idx (int): Index of the desired quantile in the rolling observation list.

    Example:

        ::

            >>> from creme import stats
            >>> import numpy as np

            >>> rolling_quantile = stats.RollingQuantile(window_size = 100,
            ...                                          quantile = 0.5)

            >>> for i in range(0,1001):
            ...     _ = rolling_quantile.update(i)
            ...     if i%100 == 0:
            ...         print(rolling_quantile.get())
            0
            50
            150
            250
            350
            450
            550
            650
            750
            850
            950

    References:

        1. `Left sorted <https://stackoverflow.com/questions/8024571/insert-an-item-into-sorted-list-in-python>`_

    """

    def __init__(self, window_size, quantile=0.5):
        super().__init__(window_size)
        self.quantile = quantile
        self.idx = int(round(self.quantile * self.window_size + 0.5)) - 1

    @property
    def name(self):
        return f'rolling_{self.window_size}_quantile'

    def update(self, x):
        self.append(x)
        return self

    def get(self):
        if len(self) < self.window_size:
            idx = int(round(self.quantile * len(self) + 0.5)) - 1
            return self[idx]
        return self[self.idx]



import collections
import math

from .. import dist

from . import base


__all__ = ['MultinomialNB']


class MultinomialNB(base.BaseNB):
    """Naive Bayes classifier for multinomial models.

    The input vector has to contain positive values, such as counts or TF-IDF values.

    This class inherits ``predict_proba_one`` from ``naive_bayes.BaseNB`` which itself inherits
    ``predict_one`` from `base.MultiClassifier`.

    Parameters:
        alpha (float): Smoothing parameter used for avoiding zero probabilities.

    Attributes:
        n (int): Number of seen observations.
        class_counts (collections.defaultdict): Number of times each class has been seen.
        term_counts (collections.defaultdict): Number of times each term has been seen.
        class_term_counts (collections.defaultdict): Number of times each term has been seen per
            class.

    Example:

        ::

            >>> import math
            >>> from creme import compose
            >>> from creme import feature_extraction
            >>> from creme import naive_bayes

            >>> docs = [
            ...     ('Chinese Beijing Chinese', 'yes'),
            ...     ('Chinese Chinese Shanghai', 'yes'),
            ...     ('Chinese Macao', 'yes'),
            ...     ('Tokyo Japan Chinese', 'no')
            ... ]
            >>> model = compose.Pipeline([
            ...     ('tokenize', feature_extraction.CountVectorizer(on='text', lowercase=False)),
            ...     ('nb', naive_bayes.MultinomialNB(alpha=1))
            ... ])
            >>> for x, y in docs:
            ...     model = model.fit_one({'text': x}, y)

            >>> model['nb'].p_class('yes')
            0.75
            >>> cp = model['nb'].p_feature_given_class
            >>> cp('Chinese', 'yes') ==  3 / 7
            True
            >>> cp('Tokyo', 'yes') ==  1 / 14
            True
            >>> cp('Japan', 'yes') ==  1 / 14
            True
            >>> cp('Chinese', 'no') ==  2 / 9
            True
            >>> cp('Tokyo', 'no') ==  2 / 9
            True
            >>> cp('Japan', 'no') ==  2 / 9
            True

            >>> new_text = 'Chinese Chinese Chinese Tokyo Japan'
            >>> tokens = model['tokenize'].transform_one({'text': new_text})
            >>> jlh = model['nb']._joint_log_likelihood(tokens)
            >>> math.exp(jlh['yes'])
            0.0003...
            >>> math.exp(jlh['no'])
            0.0001...
            >>> model.predict_one({'text': new_text})
            'yes'

    References:

        1. `Naive Bayes text classification <https://nlp.stanford.edu/IR-book/html/htmledition/naive-bayes-text-classification-1.html>`_

    """

    def __init__(self, alpha=1.0):
        self.alpha = alpha
        self.class_dist = dist.Multinomial()
        self.feature_frequencies = collections.defaultdict(collections.Counter)
        self.class_total_frequencies = collections.Counter()

    def fit_one(self, x, y):
        self.class_dist.update(y)

        for feature, frequency in x.items():
            self.feature_frequencies[feature].update({y: frequency})
            self.class_total_frequencies.update({y: frequency})

        return self

    @property
    def classes_(self):
        return list(self.class_total_frequencies.keys())

    @property
    def n_terms(self):
        return len(self.feature_frequencies)

    def p_feature_given_class(self, f, c):
        num = self.feature_frequencies[f][c] + self.alpha
        den = self.class_total_frequencies[c] + self.alpha * self.n_terms
        return num / den

    def p_class(self, c):
        return self.class_dist.pmf(c)

    def _joint_log_likelihood(self, x):
        return {
            c: math.log(self.p_class(c)) + sum(
                frequency * math.log(self.p_feature_given_class(feature, c))
                for feature, frequency in x.items()
            )
            for c in self.classes_
        }



from .. import base


__all__ = ['Blacklister']


class Blacklister(base.Transformer):
    """Subsets a set of features by applying a blacklist.

    Parameters:
        blacklist (str or list): Key(s) to discard.

    Example:

        ::

            >>> from creme import compose

            >>> x = {'a': 42, 'b': 12}
            >>> compose.Blacklister(['a', 'zoidberg']).transform_one(x)
            {'b': 12}

            >>> compose.Blacklister('b').transform_one(x)
            {'a': 42}

    """

    def __init__(self, blacklist=None):
        if blacklist is None:
            blacklist = []
        self.blacklist = set(blacklist if isinstance(blacklist, (list, tuple)) else [blacklist])

    def transform_one(self, x):
        return {i: x[i] for i in set(x.keys()) - self.blacklist}

    def __str__(self):
        return self.__class__.__name__ + f'({list(self.blacklist)})'



import collections

from .. import base
from .. import stats


__all__ = ['MinMaxScaler', 'StandardScaler']


class StandardScaler(base.Transformer):
    """Scales the data so that it has zero mean and unit variance.

    Under the hood a running mean and a running variance are maintained. The scaling is slightly
    different than when using scikit-learn but this doesn't seem to have any impact on learning
    performance.

    Attributes:
        variances (dict): Mapping between features and instances of ``stats.Variance``.
        eps (float): Used for avoiding divisions by zero.

    Example:

        ::

              >>> import creme
              >>> import numpy as np
              >>> from sklearn import preprocessing

              >>> rng = np.random.RandomState(42)
              >>> X = [{'x': v} for v in rng.uniform(low=8, high=12, size=15)]

              >>> scaler = creme.preprocessing.StandardScaler()
              >>> for x in X:
              ...     print(scaler.fit_one(x).transform_one(x))
              {'x': 0.0}
              {'x': 0.707106781053423}
              {'x': 0.1589936150008511}
              {'x': -0.27053221501993885}
              {'x': -1.3161741265511262}
              {'x': -1.0512188306232884}
              {'x': -1.1096762393237039}
              {'x': 1.0914126848882046}
              {'x': 0.31090608493934085}
              {'x': 0.5949866913888866}
              {'x': -1.3540960657565435}
              {'x': 1.2959176343737868}
              {'x': 0.8426620963810948}
              {'x': -0.8843412114527541}
              {'x': -0.9118818053170898}


              >>> X = np.array([x['x'] for x in X]).reshape(-1, 1)
              >>> preprocessing.StandardScaler().fit_transform(X)
              array([[-0.36224883],
                     [ 1.37671717],
                     [ 0.71659166],
                     [ 0.31416852],
                     [-1.02177407],
                     [-1.02184687],
                     [-1.31735428],
                     [ 1.1215704 ],
                     [ 0.32158263],
                     [ 0.64439399],
                     [-1.43053132],
                     [ 1.43465174],
                     [ 1.01975844],
                     [-0.85179183],
                     [-0.94388734]])

    """

    def __init__(self):
        self.variances = collections.defaultdict(stats.Variance)
        self.eps = 10e-10

    def fit_one(self, x, y=None):

        for i, xi in x.items():
            self.variances[i].update(xi)

        return self

    def transform_one(self, x):
        return {
            i: (xi - self.variances[i].mean.get()) / (self.variances[i].get() + self.eps) ** 0.5
            for i, xi in x.items()
        }


class MinMaxScaler(base.Transformer):
    """Scales the data to a fixed range from 0 to 1.

    Under the hood a running min and a running peak to peak (max - min) are maintained. The scaling
    is slightly different than when using scikit-learn but this doesn't seem to have any impact on
    learning performance.

    Attributes:
        min (dict): Mapping between features and instances of `stats.Min`.
        max (dict): Mapping between features and instances of `stats.Max`.
        eps (float): Used for avoiding divisions by zero.

    Example:

        ::

            >>> import creme
            >>> import numpy as np
            >>> from sklearn import preprocessing

            >>> rng = np.random.RandomState(42)
            >>> X = [{'x': v} for v in rng.uniform(low=8, high=12, size=15)]

            >>> scaler = creme.preprocessing.MinMaxScaler()
            >>> for x in X:
            ...     print(scaler.fit_one(x).transform_one(x))
            {'x': 0.0}
            {'x': 0.999999...}
            {'x': 0.620391...}
            {'x': 0.388976...}
            {'x': 0.0}
            {'x': 0.0}
            {'x': 0.0}
            {'x': 0.905293...}
            {'x': 0.608349...}
            {'x': 0.728172...}
            {'x': 0.0}
            {'x': 0.999999...}
            {'x': 0.855194...}
            {'x': 0.201990...}
            {'x': 0.169847...}

            >>> X = np.array([x['x'] for x in X]).reshape(-1, 1)
            >>> preprocessing.MinMaxScaler().fit_transform(X)
            array([[0.37284965],
                   [0.9797798 ],
                   [0.74938422],
                   [0.60893137],
                   [0.14266357],
                   [0.14263816],
                   [0.03950081],
                   [0.89072903],
                   [0.61151903],
                   [0.72418595],
                   [0.        ],
                   [1.        ],
                   [0.85519484],
                   [0.20199041],
                   [0.16984743]])

    """

    def __init__(self):
        self.min = collections.defaultdict(stats.Min)
        self.max = collections.defaultdict(stats.Max)
        self.eps = 10e-10

    def fit_one(self, x, y=None):

        for i, xi in x.items():
            self.min[i].update(xi)
            self.max[i].update(xi)

        return self

    def transform_one(self, x):
        return {
            i: (xi - self.min[i].get()) / (self.max[i].get() - self.min[i].get() + self.eps)
            for i, xi in x.items()
        }



import math

from . import base


class Max(base.Univariate):
    """Running max.

    Attributes:
        max (float): The running max.

    """

    def __init__(self):
        self.max = -math.inf

    @property
    def name(self):
        return 'max'

    def update(self, x):
        if x > self.max:
            self.max = x
        return self

    def get(self):
        return self.max



from . import variance


class SEM(variance.Variance):
    """Running standard error of the mean using Welford's algorithm.

    Parameters:
        ddof (int): Delta Degrees of Freedom. The divisor used in calculations is $n$ - ddof,
            where $n$ represents the number of seen elements.

    Attributes:
        n (int): Number of observations.

    Example:

        ::

            >>> import creme.stats

            >>> X = [3, 5, 4, 7, 10, 12]

            >>> sem = creme.stats.SEM()
            >>> for x in X:
            ...     print(sem.update(x).get())
            0.0
            1.0
            0.577350...
            0.853912...
            1.240967...
            1.447219...

    References:

        1. `Welford's online algorithm <https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Welford's_Online_algorithm>`_

    """

    @property
    def name(self):
        return 'sem'

    def get(self):
        return (super().get() / self.mean.n) ** 0.5



from . import base


def fetch_electricity(data_home=None):
    """Returns a stream containing a day of electricity prices in New South Wales.

    Parameters:
        data_home (str): The directory where you wish to store the data.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    1. `SPLICE-2 Comparative Evaluation: Electricity Pricing <http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.12.9405>`_

    """

    name = 'electricity'
    url = 'https://maxhalford.github.io/files/datasets/electricity.zip'

    return base.fetch_csv_dataset(
        data_home=data_home,
        url=url,
        name=name,
        target_name='class',
        types={
            'date': float,
            'day': int,
            'period': float,
            'nswprice': float,
            'nswdemand': float,
            'vicprice': float,
            'vicdemand': float,
            'transfer': float,
            'class': lambda x: True if x == 'UP' else False
        }
    )



import math

from .. import base
from .. import optim


__all__ = ['HedgeClassifier']


class HedgeClassifier(base.BinaryClassifier):
    """A binary classifier that uses the Hedge algorithm to mix models.

    Parameters:
        models (list): The set of models to hedge.
        weights (list): The initial weight of each model. If ``None`` then a uniform set of
            weights is assumed. This roughly translates to the prior amount of trust we have in
            each model.
        loss (optim.Loss): The loss function that has to be minimized.
        learning_rate (float): The learning rate by which the model weights are multiplied at each
            iteration.

    Example:

        ::

            >>> from creme import compose
            >>> from creme import ensemble
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     load_dataset=datasets.load_breast_cancer,
            ...     shuffle=False
            ... )
            >>> model = compose.Pipeline([
            ...     ('scale', preprocessing.StandardScaler()),
            ...     ('hedge', ensemble.HedgeClassifier(
            ...         models=[
            ...             linear_model.PAClassifier(mode=0),
            ...             linear_model.PAClassifier(mode=1),
            ...             linear_model.PAClassifier(mode=2),
            ...         ],
            ...         learning_rate=0.9
            ...     ))
            ... ])
            >>> metric = metrics.F1Score()

            >>> model_selection.online_score(X_y, model, metric)
            F1Score: 0.91372

            >>> model['hedge'].weights
            [9.992806...e-28, 0.999999..., 9.978434...e-28]

    References:

        1. `Online Learning from Experts: Weighed Majority and Hedge <https://www.shivani-agarwal.net/Teaching/E0370/Aug-2011/Lectures/20-scribe1.pdf>`_

    """

    def __init__(self, models=None, weights=None, loss=optim.LogLoss(), learning_rate=0.5):
        self.models = models
        self.weights = [1 / len(models)] * len(models) if weights is None else weights
        self.loss = loss
        self.learning_rate = learning_rate

    def fit_one(self, x, y):

        # Normalize the weights so that they sum up to 1
        total = sum(self.weights)
        for i, _ in enumerate(self.weights):
            self.weights[i] /= total

        for i, (model, weight) in enumerate(zip(self.models, self.weights)):
            y_pred = model.predict_one(x)
            loss = self.loss(y, y_pred)
            self.weights[i] *= math.exp(-self.learning_rate * loss)
            model.fit_one(x, y)

        return self

    def predict_proba_one(self, x):
        y_pred = sum(
            model.predict_proba_one(x)[True] * weight
            for model, weight in zip(self.models, self.weights)
        )
        return {False: 1 - y_pred, True: y_pred}



import collections
import statistics

from .. import stats

from . import base
from . import precision
from . import recall


__all__ = ['F1Score', 'MacroF1Score', 'MicroF1Score']


class F1Score(stats.Mean, base.BinaryClassificationMetric):
    """Binary F1 score.

    The F1 score is the harmonic mean of the precision and the recall.

    Example:

        ::

            >>> from creme import metrics
            >>> from sklearn.metrics import f1_score

            >>> y_true = [True, False, True, True, True]
            >>> y_pred = [True, True, False, True, True]

            >>> metric = metrics.F1Score()
            >>> for i, (y_t, y_p) in enumerate(zip(y_true, y_pred)):
            ...     metric = metric.update(y_t, y_p)
            ...     assert metric.get() == f1_score(y_true[:i+1], y_pred[:i+1])

            >>> metric
            F1Score: 0.75

    """

    def __init__(self):
        super().__init__()
        self.precision = precision.Precision()
        self.recall = recall.Recall()

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True

    def update(self, y_true, y_pred):
        self.precision.update(y_true, y_pred)
        self.recall.update(y_true, y_pred)
        return self

    def get(self):
        return statistics.harmonic_mean((self.precision.get(), self.recall.get()))


class MacroF1Score(base.MultiClassificationMetric):
    """Macro-average F1 score.

    The macro-average F1 score is the arithmetic average of the binary F1 scores of each label.

    Example:

        ::

            >>> from creme import metrics
            >>> from sklearn.metrics import f1_score

            >>> y_true = [0, 1, 2, 2, 2]
            >>> y_pred = [0, 0, 2, 2, 1]

            >>> metric = metrics.MacroF1Score()
            >>> for i, (y_t, y_p) in enumerate(zip(y_true, y_pred)):
            ...     metric = metric.update(y_t, y_p)
            ...     print(metric.get(), f1_score(y_true[:i+1], y_pred[:i+1], average='macro'))
            1.0 1.0
            0.333333... 0.333333...
            0.555555... 0.555555...
            0.555555... 0.555555...
            0.488888... 0.488888...

            >>> metric
            MacroF1Score: 0.488889

    """

    def __init__(self):
        self.f1_scores = collections.defaultdict(F1Score)
        self.classes = set()

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True

    def update(self, y_true, y_pred):
        self.classes.update({y_true, y_pred})
        for c in self.classes:
            self.f1_scores[c].update(y_true == c, y_pred == c)
        return self

    def get(self):
        total = sum(f1.get() for f1 in self.f1_scores.values())
        if total == 0.:
            return 0.
        return total / len(self.f1_scores)


class MicroF1Score(precision.MicroPrecision):
    """Micro-average F1 score.

    The micro-average F1 score is exactly equivalent to the micro-average precision as well as the
    micro-average recall score.

    Example:

        ::

            >>> from creme import metrics
            >>> from sklearn.metrics import f1_score

            >>> y_true = [0, 1, 2, 2, 2]
            >>> y_pred = [0, 0, 2, 2, 1]

            >>> metric = metrics.MicroF1Score()
            >>> for i, (y_t, y_p) in enumerate(zip(y_true, y_pred)):
            ...     metric = metric.update(y_t, y_p)
            ...     print(metric.get(), f1_score(y_true[:i+1], y_pred[:i+1], average='micro'))
            1.0 1.0
            0.5 0.5
            0.666666... 0.666666...
            0.75 0.75
            0.6 0.6

            >>> metric
            MicroF1Score: 0.6

    References:

        1. `Why are precision, recall and F1 score equal when using micro averaging in a multi-class problem? <https://simonhessner.de/why-are-precision-recall-and-f1-score-equal-when-using-micro-averaging-in-a-multi-class-problem/>`_

    """



import collections

import numpy as np
from sklearn import utils

from .. import base


__all__ = ['KMeans']


def euclidean_distance(a, b):
    return sum((a.get(k, 0) - b.get(k, 0)) ** 2 for k in set([*a.keys(), *b.keys()]))


class KMeans(base.Clusterer):
    """Incremental k-means.

    The most common way to implement batch k-means is to use Lloyd's algorithm, which consists in
    assigning all the data points to a set of cluster centers and then moving the centers
    accordingly. This requires multiple passes over the data and thus isn't applicable in a
    streaming setting.

    In this implementation we start by finding the cluster that is closest to the current
    observation. We then move the cluster's central position towards the new observation. The
    ``halflife`` parameter determines by how much to move the cluster toward the new observation.
    You will get better results if you scale your data appropriately.

    Parameters:
        n_clusters (int): Maximum number of clusters to assign.
        halflife (float): Amount by which to move the cluster centers, a reasonable value if
            between 0 and 1.
        mu (float): Mean of the normal distribution used to instantiate cluster positions.
        sigma (float): Standard deviation of the normal distribution used to instantiate cluster
            positions.
        distance (callable): Metric used to compute distances between an observation and a cluster.
        random_state (int, RandomState instance or None, default=None): If int, ``random_state`` is
            the seed used by the random number generator; if ``RandomState`` instance,
            ``random_state`` is the random number generator; if ``None``, the random number
            generator is the ``RandomState`` instance used by ``np.random``.

    Attributes:
        centers (dict): Central positions of each cluster.

    Example:

        In the following example the cluster assignments are exactly the same as when using
        ``sklearn``'s batch implementation. However changing the ``halflife`` parameter will
        produce different outputs.

        ::

            >>> from creme import cluster
            >>> from creme import compat
            >>> import numpy as np
            >>> X = np.array([[1, 2], [1, 4], [1, 0],
            ...               [4, 2], [4, 4], [4, 0]])
            >>> k_means = cluster.KMeans(n_clusters=2, halflife=0.4, sigma=3, random_state=42)
            >>> k_means = compat.SKLClustererWrapper(k_means)
            >>> k_means = k_means.fit(X)

            >>> k_means.predict(X)
            array([0, 0, 0, 1, 1, 1], dtype=int32)

            >>> k_means.predict([[0, 0], [4, 4]])
            array([0, 1], dtype=int32)

    References:

        1. `Sequential k-Means Clustering <http://www.cs.princeton.edu/courses/archive/fall08/cos436/Duda/C/sk_means.htm>`_

    """

    def __init__(self, n_clusters=8, halflife=0.5, mu=0, sigma=1, distance=euclidean_distance,
                 random_state=None):
        self.n_clusters = n_clusters
        self.halflife = halflife
        self.mu = mu
        self.sigma = sigma
        self.distance = distance
        self.random_state = utils.check_random_state(random_state)
        self.centers = {
            i: collections.defaultdict(self.random_normal)
            for i in range(n_clusters)
        }

    def random_normal(self):
        """Returns a random value sampled from a normal distribution."""
        return self.random_state.normal(self.mu, self.sigma)

    @property
    def cluster_centers_(self):
        """Returns the cluster centers in the same format as scikit-learn."""
        return np.array([
            list(coords.values())
            for coords in self.centers.values()
        ])

    def fit_one(self, x, y=None):

        # Find the cluster with the closest center
        closest = self.predict_one(x)

        # Move the cluster's center
        for i, xi in x.items():
            self.centers[closest][i] += self.halflife * (xi - self.centers[closest][i])

        return self

    def predict_one(self, x):
        return min(self.centers, key=lambda c: self.distance(x, self.centers[c]))



"""
Module for computing running statistics
"""
from .base import Bivariate
from .base import Univariate
from .count import Count
from .covariance import Covariance
from .entropy import Entropy
from .ewmean import EWMean
from .kurtosis import Kurtosis
from .max import Max
from .mean import BayesianMean
from .mean import Mean
from .min import Min
from .mode import Mode
from .n_unique import NUnique
from .pearson import PearsonCorrelation
from .ptp import PeakToPeak
from .quantile import Quantile
from .rolling_max import RollingMax
from .rolling_mean import RollingMean
from .rolling_min import RollingMin
from .rolling_mode import RollingMode
from .rolling_ptp import RollingPeakToPeak
from .rolling_quantile import RollingQuantile
from .rolling_sem import RollingSEM
from .rolling_sum import RollingSum
from .rolling_variance import RollingVariance
from .sem import SEM
from .skew import Skew
from .sum import Sum
from .variance import Variance


__all__ = [
    'BayesianMean',
    'Bivariate',
    'Count',
    'Covariance',
    'Entropy',
    'EWMean',
    'Kurtosis',
    'Max',
    'Mean',
    'Min',
    'Mode',
    'NUnique',
    'PeakToPeak',
    'PearsonCorrelation',
    'Quantile',
    'RollingMean',
    'RollingQuantile',
    'RollingMax',
    'RollingMin',
    'RollingMode',
    'RollingPeakToPeak',
    'RollingSEM',
    'RollingSum',
    'RollingVariance',
    'SEM',
    'Skew',
    'Sum',
    'Univariate',
    'Variance'
]



from . import rolling_sum


class RollingMean(rolling_sum.RollingSum):
    """Running average over a window.

    Parameters:
        window_size (int): Size of the rolling window.

    Example:

        ::

            >>> import creme

            >>> X = [1, 2, 3, 4, 5, 6]

            >>> rolling_mean = creme.stats.RollingMean(window_size=2)
            >>> for x in X:
            ...     print(rolling_mean.update(x).get())
            1.0
            1.5
            2.5
            3.5
            4.5
            5.5

            >>> rolling_mean = creme.stats.RollingMean(window_size=3)
            >>> for x in X:
            ...     print(rolling_mean.update(x).get())
            1.0
            1.5
            2.0
            3.0
            4.0
            5.0

    """

    @property
    def name(self):
        return f'rolling_{self.window_size}_mean'

    def get(self):
        return super().get() / len(self) if len(self) > 0 else 0



import abc
import typing

from . import lr_schedule


class Optimizer(abc.ABC):

    def __init__(self, lr: typing.Union[lr_schedule.LRScheduler, float]):
        self.lr = lr_schedule.ConstantLR(lr) if isinstance(lr, float) else lr
        self.n_iterations = 0

    @property
    def learning_rate(self) -> float:
        return self.lr.get(self.n_iterations)

    def update_before_pred(self, w: dict) -> dict:
        return w

    def _update_after_pred(self, w: dict, g: dict) -> dict:
        raise NotImplementedError

    def update_after_pred(self, w: dict, g: dict):

        # Update the weights
        w = self._update_after_pred(w, g)

        # Update the iteration counter
        self.n_iterations += 1

        return w



"""
Utilities for measuring the performance of online learning algorithms.
"""
from . import base


__all__ = ['online_score']


def online_score(X_y, model, metric):
    """Computes the online score of a model given a stream of data and a metric.

    Parameters:
        X_y (generator): A stream of (features, target) tuples.
        model (estimator)
        metric (callable)

    Returns:
        metric

    """

    # Check that the model and the metric are in accordance
    if not metric.works_with(model):
        raise ValueError(f"{metric.__class__.__name__} metric can't be used to evaluate a " +
                         f'{model.__class__.__name__}')

    # Determine if predict_one or predict_proba_one should be used
    pred_func = model.predict_one
    if isinstance(model, base.Classifier) and not metric.requires_labels:
        pred_func = model.predict_proba_one

    # Train the model and use the out-of-fold predictions to update the metric
    for x, y in X_y:
        y_pred = pred_func(x)
        model = model.fit_one(x, y)
        if y_pred != {} and y_pred is not None:
            metric.update(y, y_pred)

    return metric



import collections

from . import base


class Multinomial(base.DiscreteDistribution):
    """Multinomial distribution.

    Example:

        ::

            >>> d1 = Multinomial().update('sunny')
            >>> d1.mode()
            'sunny'

            >>> d2 = Multinomial().update('rainy').update('rainy')
            >>> d2.mode()
            'rainy'

            >>> d1 += d2
            >>> d1.mode()
            'rainy'

            >>> d1.pmf('cloudy')
            0.0

    """

    def __init__(self):
        self.counter = collections.Counter()
        self.n = 0

    def update(self, x):
        self.counter.update([x])
        self.n += 1
        return self

    def mode(self):
        if self.counter:
            return max(self.counter, key=self.counter.get)
        return None

    def pmf(self, x):
        return self.counter.get(x, 0) / self.n

    def __iadd__(self, other):
        self.counter.update(other.counter)
        self.n += other.n
        return self

    def __imul__(self, constant):
        for i in self.counter:
            self.counter[i] *= constant
        return self

    def __copy__(self):
        """Custom copying to make ``copy.copy`` faster."""
        clone = Multinomial()
        clone.counter = collections.Counter(self.counter)
        clone.n = self.n
        return clone

    def __deepcopy__(self, memo):
        """Custom copying to make ``copy.deepcopy`` faster."""
        return self.__copy__()



from . import rolling_variance


class RollingSEM(rolling_variance.RollingVariance):
    """Running standard error of the mean over a window.

    Parameters:
        window_size (int): Size of the rolling window.
        ddof (int): Delta Degrees of Freedom for the variance.

    Example:

        ::
            >>> import creme

            >>> X = [1, 4, 2, -4, -8, 0]

            >>> rolling_sem = creme.stats.RollingSEM(ddof=1, window_size=2)
            >>> for x in X:
            ...     print(rolling_sem.update(x).get())
            0.0
            1.5
            1.0
            3.0
            2.0
            4.0

            >>> rolling_sem = creme.stats.RollingSEM(ddof=1, window_size=3)
            >>> for x in X:
            ...     print(rolling_sem.update(x).get())
            0.0
            1.5
            0.881917...
            2.403700...
            2.905932...
            2.309401...

    """

    @property
    def name(self):
        return f'rolling_{self.window_size}_sem'

    def get(self):
        return (super().get() / len(self.rolling_mean)) ** 0.5



"""
Generalized linear models optimized with online gradient descent using the
:mod:`creme.optim` module.
"""
from .fm import FMRegressor
from .lin_reg import LinearRegression
from .log_reg import LogisticRegression
from .pa import PAClassifier
from .pa import PARegressor
from .softmax import SoftmaxRegression


__all__ = [
    'FMRegressor',
    'LinearRegression',
    'LogisticRegression',
    'PAClassifier',
    'PARegressor',
    'SoftmaxRegression'
]



from . import base
from . import max
from . import min


class PeakToPeak(base.Univariate):
    """Running peak to peak (max - min).

    Attributes:
        max (stats.Max): The running max.
        min (stats.Min): The running min.
        p2p (float): The running peak to peak.

    Example:

        ::

            >>> from creme import stats

            >>> X = [1, -4, 3, -2, 2, 4]
            >>> ptp = stats.PeakToPeak()
            >>> for x in X:
            ...     print(ptp.update(x).get())
            0
            5
            7
            7
            7
            8

    """

    def __init__(self):
        self.max = max.Max()
        self.min = min.Min()

    @property
    def name(self):
        return 'ptp'

    def update(self, x):
        self.max.update(x)
        self.min.update(x)
        return self

    def get(self):
        return self.max.get() - self.min.get()



from . import base
from . import rolling_max
from . import rolling_min


class RollingPeakToPeak(base.Univariate):
    """Running peak to peak (max - min) over a window.

    Parameters:
        window_size (int): Size of the rolling window.

    Attributes:
        max (stars.RollingMax): The running rolling max.
        min (stars.RollingMin): The running rolling min.

    Example:

        ::

            >>> from creme import stats

            >>> X = [1, -4, 3, -2, 2, 1]
            >>> ptp = stats.RollingPeakToPeak(window_size=2)
            >>> for x in X:
            ...     print(ptp.update(x).get())
            0
            5
            7
            5
            4
            1

    """

    def __init__(self, window_size):
        self.window_size = window_size
        self.max = rolling_max.RollingMax(window_size)
        self.min = rolling_min.RollingMin(window_size)

    @property
    def name(self):
        return f'rolling_{self.window_size}_ptp'

    def update(self, x):
        self.max.update(x)
        self.min.update(x)
        return self

    def get(self):
        return self.max.get() - self.min.get()



from . import base
from . import mean


class Variance(base.Univariate):
    """Running variance using Welford's algorithm.

    Parameters:
        ddof (int): Delta Degrees of Freedom. The divisor used in calculations is $n$ - ddof,
            where $n$ represents the number of seen elements.

    Attributes:
        mean (stats.Mean): The running mean.
        sos (float): The running sum of squares.

    Example:

        ::

            >>> import creme.stats

            >>> X = [3, 5, 4, 7, 10, 12]

            >>> var = creme.stats.Variance()
            >>> for x in X:
            ...     print(var.update(x).get())
            0.0
            2.0
            1.0
            2.916666...
            7.7
            12.56666...

    References:

        1. `Welford's online algorithm <https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Welford's_Online_algorithm>`_

    """

    def __init__(self, ddof=1):
        self.ddof = ddof
        self.mean = mean.Mean()
        self.sos = 0

    @property
    def name(self):
        return 'variance'

    @property
    def n(self):
        return self.mean.n

    def update(self, x):
        self.sos += (x - self.mean.get()) * (x - self.mean.update(x).get())
        return self

    def get(self):
        return self.sos / max(1, self.mean.n - self.ddof)



import copy

from .. import base


__all__ = ['SplitRegressor']


class SplitRegressor(base.Regressor):
    """Runs a different regressor based on the value of a specified attribute.

    Parameters:
        on (str): The feature on which to perform the split.
        models (dict): A mapping between feature values and regressor.
        default_model (base.Regressor): The regressor used for feature values that are not
            specified in ``models``.

    Example:

        ::

            >>> from creme import compose
            >>> from creme import dummy
            >>> from creme import stats

            >>> X = [
            ...     {'key': 'a', 'y': 2},
            ...     {'key': 'a', 'y': 3},
            ...     {'key': 'a', 'y': 4},
            ...     {'key': 'b', 'y': 1},
            ...     {'key': 'b', 'y': 42},
            ...     {'key': 'b', 'y': 1337},
            ...     {'key': 'c', 'y': 6},
            ...     {'key': 'c', 'y': 1},
            ...     {'key': 'c', 'y': 6}
            ... ]

            >>> model = compose.SplitRegressor(
            ...     on='key',
            ...     models={
            ...         'a': dummy.StatisticRegressor(stats.Mean()),
            ...         'b': dummy.StatisticRegressor(stats.Quantile(0.5))
            ...     },
            ...     default_model=dummy.StatisticRegressor(stats.Min())
            ... )

            >>> for x in X:
            ...     y = x.pop('y')
            ...     model = model.fit_one(x, y)

            >>> model.models['a'].statistic.get()
            3.0

            >>> model.predict_one({'key': 'a'})
            3.0

            >>> model.models['b'].statistic.get()
            42

            >>> model.default_model.statistic.get()
            1

    """

    def __init__(self, on, models, default_model):
        self.on = on
        self.models = copy.deepcopy(models)
        self.default_model = copy.deepcopy(default_model)

    def fit_one(self, x, y):
        x = copy.copy(x)
        key = x[self.on]
        x.pop(self.on)

        self.models.get(key, self.default_model).fit_one(x, y)
        return self

    def predict_one(self, x):
        x = copy.copy(x)
        key = x[self.on]
        x.pop(self.on)

        return self.models.get(key, self.default_model).predict_one(x)



from .sklearn import convert_creme_to_sklearn
from .sklearn import convert_sklearn_to_creme
from .sklearn import CremeClassifierWrapper
from .sklearn import CremeRegressorWrapper
from .sklearn import SKLRegressorWrapper
from .sklearn import SKLClassifierWrapper
from .sklearn import SKLClustererWrapper
from .sklearn import SKLTransformerWrapper



import abc
import collections
import functools
import math

from .. import base


def entropy(counts):
    """Computes entropy using counts.

    Parameters:
        counts (collections.Counter)

    References:

        1. `Calculating entropy <https://www.johndcook.com/blog/2013/08/17/calculating-entropy/>`_

    """
    N = sum(counts.values())
    return math.log2(N) - sum(n * math.log2(n) for n in counts.values()) / N


class Node(abc.ABC):

    @abc.abstractmethod
    def get_leaf(self, x):
        """Recursively search for the leaf that contains ``x``."""


class Branch(Node):

    def __init__(self, splitter, left, right):
        self.splitter = splitter
        self.left = left
        self.right = right

    def get_leaf(self, x):
        if self.splitter(x):
            return self.left.next()
        return self.right.next()


class Leaf(Node):

    def __init__(self):
        self.class_counts = collections.Counter()
        dd = collections.defaultdict
        self.feature_counts = dd(functools.partial(dd, collections.Counter))

    def get_leaf(self, x):
        return self

    @property
    def is_pure(self):
        return len(self.class_counts) <= 1


class DecisionTreeClassifier(base.MultiClassifier):

    def __init__(self):
        self.root = Leaf()

    def fit_one(self, x, y):

        # Assign the observation to a leaf
        leaf = self.root.get_leaf(x)

        # Update the leaf's overall class counts
        leaf.class_counts.update((y,))

        # Update the leaf's class counts per feature value
        for i, xi in x.items():
            leaf.feature_counts[i][xi].update((y,))

        if not leaf.is_pure:

            leaf_entropy = entropy(leaf.class_counts)

            for i in leaf.feature_counts:
                pass



from .estimator_checks import check_estimator
from .histogram import Histogram
from .math import chain_dot
from .math import clip
from .math import dot
from .math import norm
from .math import prod
from .math import sigmoid
from .math import softmax
from .window import Window
from .window import SortedWindow


__all__ = [
    'chain_dot',
    'check_estimator',
    'clip',
    'dot',
    'Histogram',
    'norm',
    'prod',
    'sigmoid',
    'softmax',
    'SortedWindow',
    'Window'
]



from . import base


def fetch_restaurants(data_home=None):
    """Returns a stream containing the data from the Kaggle Recruit Restaurants challenge.

    The data is ordered by data and then by store ID.

    Parameters:
        data_home (str): The directory where you wish to store the data.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    1. `Recruit Restaurant Visitor Forecasting <https://www.kaggle.com/c/recruit-restaurant-visitor-forecasting>`_

    """

    name = 'kaggle_recruit_restaurants'
    url = 'https://maxhalford.github.io/files/datasets/kaggle_recruit_restaurants.zip'

    return base.fetch_csv_dataset(
        data_home=data_home,
        url=url,
        name=name,
        target_name='visitors',
        types={
            'latitude': float,
            'longitude': float,
            'visitors': int,
            'is_holiday': bool
        },
        parse_dates={'date': '%Y-%m-%d'}
    )



"""
Module for preparing data to make it compatible with certain machine learning algorithms.
"""
from .feature_hasher import FeatureHasher
from .one_hot import OneHotEncoder
from .poly import PolynomialExtender
from .scale import StandardScaler
from .scale import MinMaxScaler


__all__ = [
    'FeatureHasher',
    'MinMaxScaler',
    'OneHotEncoder',
    'PolynomialExtender',
    'StandardScaler'
]



import collections
import itertools

try:
    import graphviz
    GRAPHVIZ_INSTALLED = True
except ImportError:
    GRAPHVIZ_INSTALLED = False

from sklearn.utils import metaestimators

from . import func
from . import union


__all__ = ['Pipeline']


class Pipeline(collections.OrderedDict):
    """Chains a sequence of estimators.

    Sequentially apply a list of estimators. Pipelines helps to define machine learning systems in a
    declarative style, which makes a lot of sense when we think in a stream manner. For further
    information and practical examples, take a look at the
    `user guide <../notebooks/the-art-of-using-pipelines.html>`_.

    Parameters:
        steps (list): Ideally a list of (name, estimator) tuples. If an estimator is given without
            a name then a name is automatically inferred from the estimator.

    Example:

        ::

            >>> from creme import compose
            >>> from creme import feature_extraction
            >>> from creme import linear_model
            >>> from creme import preprocessing

            >>> tfidf = feature_extraction.TFIDFVectorizer('text')
            >>> counts = feature_extraction.CountVectorizer('text')
            >>> text_part = compose.Whitelister('text') | (tfidf + counts)

            >>> num_part = compose.Whitelister(['a', 'b']) | preprocessing.PolynomialExtender()

            >>> model = text_part + num_part
            >>> model |= preprocessing.StandardScaler()
            >>> model |= linear_model.LinearRegression()

            >>> dot = model.draw()

        .. image:: ../_static/pipeline_docstring.svg
            :align: center

    """

    def __init__(self, steps=None):
        if steps is not None:
            for step in steps:
                self |= step

    def __or__(self, other):
        """Inserts a step at the end of the pipeline."""
        self.add_step(other, at_start=False)
        return self

    def __ror__(self, other):
        """Inserts a step at the start of the pipeline."""
        self.add_step(other, at_start=True)
        return self

    def __add__(self, other):
        """Merges with another Pipeline or TransformerUnion into a TransformerUnion."""
        if isinstance(other, union.TransformerUnion):
            return other.__add__(self)
        return union.TransformerUnion([self, other])

    def __radd__(self, other):
        """Merges with another Pipeline or TransformerUnion into a TransformerUnion."""
        if isinstance(other, union.TransformerUnion):
            return other.__add__(self)
        return union.TransformerUnion([other, self])

    def __str__(self):
        """Return a human friendly representation of the pipeline."""
        return ' | '.join(self.keys())

    @property
    def __class__(self):
        """Returns the class of the final estimator for type checking purposes.

        A Pipeline is semantically equivalent to it's final estimator in terms of usage. This is
        mostly used for deceiving the ``isinstance`` method.

        """
        return self.final_estimator.__class__

    def add_step(self, step, at_start):
        """Adds a step to either end of the pipeline while taking care of the input type."""

        # Infer a name if none is given
        if not isinstance(step, (list, tuple)):
            step = (str(step), step)

        # If a function is given then wrap it in a FuncTransformer
        if callable(step[1]):
            step = (step[1].__class__, func.FuncTransformer(step[1]))

        # Prefer clarity to magic
        if step[0] in self:
            raise KeyError(f'{step[0]} already exists')

        # Store the step
        self[step[0]] = step[1]

        # Move the step to the start of the pipeline if so instructed
        if at_start:
            self.move_to_end(step[0], last=False)

    @property
    def final_estimator(self):
        """The final estimator."""
        return self[next(reversed(self))]

    def fit_one(self, x, y=None):
        """Fits each step with ``x``."""
        x_transformed = x

        for transformer in itertools.islice(self.values(), len(self) - 1):
            x_transformed = transformer.transform_one(x_transformed)

            # The supervised parts of a TransformerUnion have to be updated
            if isinstance(transformer, union.TransformerUnion):
                for sub_transformer in transformer.values():
                    if sub_transformer.is_supervised:
                        sub_transformer.fit_one(x, y)
                continue

            # If a transformer is supervised then it has to be updated
            if transformer.is_supervised:
                transformer.fit_one(x, y)

        self.final_estimator.fit_one(x_transformed, y)
        return self

    def run_transformers(self, x):
        for transformer in itertools.islice(self.values(), len(self) - 1):

            if isinstance(transformer, union.TransformerUnion):
                for sub_transformer in transformer.values():
                    if not sub_transformer.is_supervised:
                        sub_transformer.fit_one(x)
                x = transformer.transform_one(x)
                continue

            if not transformer.is_supervised:
                transformer.fit_one(x)
            x = transformer.transform_one(x)

        return x

    @metaestimators.if_delegate_has_method(delegate='final_estimator')
    def transform_one(self, x):
        """Transform an input.

        Only works if each estimator has a ``transform_one`` method.

        """
        x = self.run_transformers(x)
        final_tranformer = self.final_estimator
        if not final_tranformer.is_supervised:
            final_tranformer.fit_one(x)
        x = final_tranformer.transform_one(x)
        return x

    @metaestimators.if_delegate_has_method(delegate='final_estimator')
    def predict_one(self, x):
        """Predict output.

        Only works if each estimator has a ``transform_one`` method and the final estimator has a
        ``predict_one`` method.

        """
        x = self.run_transformers(x)
        return self.final_estimator.predict_one(x)

    @metaestimators.if_delegate_has_method(delegate='final_estimator')
    def predict_proba_one(self, x):
        """Predicts probabilities.

        Only works if each estimator has a ``transform_one`` method and the final estimator has a
        ``predict_proba_one`` method.

        """
        x = self.run_transformers(x)
        return self.final_estimator.predict_proba_one(x)

    def draw(self):
        """Draws the pipeline using the ``graphviz`` library."""

        if not GRAPHVIZ_INSTALLED:
            raise ImportError('graphviz is not installed')

        def get_first_estimator(d):
            """Gets first estimator key of a Pipeline or TransformerUnion."""

            for first_key in d.keys():
                first_step = d.get(first_key)
                break

            if isinstance(first_step, (Pipeline, union.TransformerUnion)):
                # Recurse
                first_key = get_first_estimator(first_step)

            return first_key

        def draw_step(node, previous_node):
            """Draws a node and its previous edge."""
            if node in nodes:
                node = node + '_'
                return draw_step(node, previous_node)

            graph.node(node, node.rstrip('_'))
            graph.edge(previous_node, node)
            nodes.append(node)
            edges.append(previous_node)

        def draw_steps(d=self, skip_first=False):
            """Draws all estimators graph nodes and edges."""

            union_ending_node_ix = None

            for key in list(d.keys()):

                if skip_first:
                    skip_first = False
                    continue

                step = d.get(key)

                # If step is a Pipeline recurse on step
                if isinstance(step, Pipeline):
                    draw_steps(step)

                # If step is a TransformerUnion, dive inside
                elif isinstance(step, union.TransformerUnion):

                    node_before_union = nodes[-1]

                    # Draw each TransformerUnion steps
                    for ix, sub_key in enumerate(step.keys()):

                        sub_step = step.get(sub_key)

                        # If sub step is another nested step, draw its first estimator and recurse
                        if isinstance(sub_step, (Pipeline, union.TransformerUnion)):
                            sub_sub_key = get_first_estimator(sub_step)
                            draw_step(node=sub_sub_key, previous_node=node_before_union)
                            draw_steps(d=sub_step, skip_first=True)
                        # Else just draw it
                        else:
                            draw_step(node=sub_key, previous_node=nodes[-1 - ix])

                    union_ending_node_ix = len(nodes)

                else:
                    draw_step(key, nodes[-1])

                # If previous step was a TransformerUnion and following node have been drawn
                if union_ending_node_ix == len(nodes) - 1:
                    # Connect TransformerUnion child nodes with the next step
                    for node in nodes[1: -1]:
                        if node not in edges:
                            graph.edge(node, nodes[union_ending_node_ix])
                            edges.append(node)
                    # Reset TransformerUnion flag
                    union_ending_node_ix = None

        nodes, edges = ['input'], []
        graph = graphviz.Digraph()
        graph.node('input')

        draw_steps()

        graph.node('output')
        graph.edge(nodes[-1], 'output')

        return graph



import abc


class Univariate(abc.ABC):

    @property
    @abc.abstractmethod
    def name(self):
        """The name of the statistic is used for naming automatically generated features."""

    @abc.abstractmethod
    def update(self, x: float):
        """Updates and returns the called instance."""

    @abc.abstractmethod
    def get(self) -> float:
        """Returns the current value."""

    def __str__(self):
        return f'{self.__class__.__name__}: {self.get():.6f}'.rstrip('0')


class Bivariate(abc.ABC):

    @abc.abstractmethod
    def update(self, x, y):
        """Updates and returns the called instance."""

    @abc.abstractmethod
    def get(self) -> float:
        """Returns the current value."""

    def __str__(self):
        return f'{self.__class__.__name__}: {self.get():.6f}'.rstrip('0')



import os

from .. import stream


def load_airline():
    """Returns a stream of monthly number of international airline passengers.

    The data is ordered by month.

    Yields:
        tuple: A pair (``x``, ``y``) where ``x`` is a dict of features and ``y`` is the target.

    1. `International airline passengers: monthly totals in thousands. Jan 49 – Dec 60 <https://datamarket.com/data/set/22u3/international-airline-passengers-monthly-totals-in-thousands-jan-49-dec-60#!ds=22u3&display=line>`_

    """
    return stream.iter_csv(
        os.path.join(os.path.dirname(__file__), 'data', 'airline-passengers.csv'),
        target_name='passengers',
        types={'passengers': int},
        parse_dates={'month': '%Y-%m'}
    )



from . import base
from . import rolling_mean


class RollingVariance(base.Univariate):
    """Running variance over a window.

    Parameters:
        window_size (int): Size of the rolling window.
        ddof (int): Delta Degrees of Freedom for the variance.

    Attributes:
        sos (float): Sum of squares.
        rolling_mean (stats.RollingMean): The running rolling mean.

    Example:

        ::

            >>> import creme

            >>> X = [1, 4, 2, -4, -8, 0]

            >>> rolling_variance = creme.stats.RollingVariance(ddof=1, window_size=2)
            >>> for x in X:
            ...     print(rolling_variance.update(x).get())
            0.0
            4.5
            2.0
            18.0
            8.0
            32.0

            >>> rolling_variance = creme.stats.RollingVariance(ddof=1, window_size=3)
            >>> for x in X:
            ...     print(rolling_variance.update(x).get())
            0.0
            4.5
            2.333333...
            17.333333...
            25.333333...
            16.0

    """

    def __init__(self, window_size, ddof=1):
        self.window_size = window_size
        self.ddof = ddof
        self.sos = 0
        self.rolling_mean = rolling_mean.RollingMean(window_size)

    @property
    def name(self):
        return f'rolling_{self.window_size}_variance'

    def update(self, x):
        if len(self.rolling_mean) >= self.window_size:
            self.sos -= self.rolling_mean[0] ** 2

        self.sos += x * x
        self.rolling_mean.update(x)
        return self

    @property
    def correction_factor(self):
        n = len(self.rolling_mean)
        if n > self.ddof:
            return n / (n - self.ddof)
        return 1

    def get(self):
        variance = (self.sos / len(self.rolling_mean)) - self.rolling_mean.get() ** 2
        return self.correction_factor * variance



"""
A set of sequential optimizers and learning rate schedulers. Also contains loss functions commonly
used in machine learning.
"""
from .ada_delta import AdaDelta
from .ada_grad import AdaGrad
from .adam import Adam
from .base import Optimizer
from .ftrl import FTRLProximal
from .losses import AbsoluteLoss
from .losses import BinaryClassificationLoss
from .losses import CauchyLoss
from .losses import CrossEntropy
from .losses import EpsilonInsensitiveHingeLoss
from .losses import HingeLoss
from .losses import LogLoss
from .losses import MultiClassificationLoss
from .losses import RegressionLoss
from .losses import SquaredLoss
from .lr_schedule import ConstantLR
from .lr_schedule import InverseScalingLR
from .lr_schedule import OptimalLR
from .momentum import Momentum
from .nesterov import NesterovMomentum
from .rms_prop import RMSProp
from .vanilla_sgd import VanillaSGD


__all__ = [
    'AbsoluteLoss',
    'AdaDelta',
    'AdaGrad',
    'Adam',
    'BinaryClassificationLoss',
    'CauchyLoss',
    'ConstantLR',
    'CrossEntropy',
    'EpsilonInsensitiveHingeLoss',
    'FTRLProximal',
    'HingeLoss',
    'InverseScalingLR',
    'LogLoss',
    'Loss',
    'Momentum',
    'MultiClassificationLoss',
    'NesterovMomentum',
    'OptimalLR',
    'Optimizer',
    'RegressionLoss',
    'RMSProp',
    'SquaredLoss',
    'VanillaSGD'
]



import bisect
import collections
import math


def find_index(seq, val):
    """Assume ``seq`` is sorted and search for the position of ``val``.

    References:

        1. `Searching Sorted Lists <https://docs.python.org/3/library/bisect.html#searching-sorted-lists>`_

    """
    i = bisect.bisect_left(seq, val)
    if i != len(seq) and seq[i] == val:
        return i
    raise ValueError


class Histogram(collections.Counter):
    """Streaming histogram data structure.

    Parameters:
        max_bins (int): Maximal number of bins.

    Attributes:
        n_bins (int): Current number of bins.
        sorted_bins (list): Sorted left bin values.

    Example:

        ::

            >>> from creme import utils
            >>> import matplotlib.pyplot as plt
            >>> import numpy as np

            >>> np.random.seed(42)

            >>> values = np.hstack((
            ...     np.random.normal(-3, 1, 1000),
            ...     np.random.normal(3, 1, 1000),
            ... ))

            >>> hist = utils.Histogram(max_bins=12)

            >>> for x in values:
            ...     hist = hist.update(x)

            >>> print(hist)
            [-5.53514, -4.55869): 9
            [-4.55869, -3.61492): 94
            [-3.61492, -2.58699): 338
            [-2.58699, -1.44212): 423
            [-1.44212, -0.12576): 128
            [-0.12576, 0.98003): 13
            [0.98003, 1.81977): 31
            [1.81977, 2.69824): 150
            [2.69824, 3.57935): 339
            [3.57935, 4.58907): 329
            [4.58907, 5.89119): 142

            >>> ax = plt.bar(hist.keys(), hist.values())

        .. image:: ../_static/histogram_docstring.svg
            :align: center

    References:

        1. `A Streaming Parallel Decision Tree Algorithm <http://jmlr.org/papers/volume11/ben-haim10a/ben-haim10a.pdf>`_
        2. `Streaming Approximate Histograms in Go <https://www.vividcortex.com/blog/2013/07/08/streaming-approximate-histograms/>`_
        3. `Go implementation <https://github.com/VividCortex/gohistogram>`_

    """

    def __init__(self, max_bins=256):
        self.sorted_bins = []
        self.max_bins = max_bins

    @property
    def n_bins(self):
        return len(self)

    def update(self, x):

        super().update([x])
        bisect.insort_left(self.sorted_bins, x)

        # Bins have to be merged if there are more than max_bins
        while self.n_bins > self.max_bins:

            # Find the nearest bins
            min_val = math.inf
            min_idx = None

            # TODO: this loop should be Cythonized
            for i, (a, b) in enumerate(zip(self.sorted_bins[:-1], self.sorted_bins[1:])):
                if b - a < min_val:
                    min_val = b - a
                    min_idx = i

            left_bin = self.sorted_bins[min_idx]
            right_bin = self.sorted_bins[min_idx + 1]

            # Merge the two bins by summing their count and making a weighted average for each one
            total_count = self[left_bin] + self[right_bin]

            new_bin = left_bin * self[left_bin] + right_bin * self[right_bin]
            new_bin /= total_count
            bisect.insort_left(self.sorted_bins, new_bin)

            # Deletion of the two bins to replace them with the new one resulting from their merging
            for k in [right_bin, left_bin]:
                self.pop(k)
                self.sorted_bins.pop((find_index(self.sorted_bins, k)))

            super().update({new_bin: total_count})

        return self

    def __str__(self):
        return '\n'.join(
            f'[{a:.5f}, {b:.5f}): {self[a]}'
            for a, b in zip(self.sorted_bins[:-1], self.sorted_bins[1:])
        )



from . import base


class Sum(base.Univariate):
    """Running sum.

    Attributes:
        sum (float) : The running sum.

    Example:

        ::

            >>> from creme import stats

            >>> X = [-5, -3, -1, 1, 3, 5]
            >>> mean = stats.Sum()
            >>> for x in X:
            ...     print(mean.update(x).get())
            -5.0
            -8.0
            -9.0
            -8.0
            -5.0
            0.0

    """

    def __init__(self):
        self.sum = 0.

    @property
    def name(self):
        return 'sum'

    def update(self, x):
        self.sum += x
        return self

    def get(self):
        return self.sum



"""
Models for predicting ratings from (user, item) pairs.
"""

from .normal import RandomNormal
from .sgd_baseline import SGDBaseline
from .svd import SVD


__all__ = ['RandomNormal', 'SGDBaseline', 'SVD']



import collections

from . import base


__all__ = ['Mode']


class Mode(base.Univariate):
    """Running mode.

    The mode is simply the most common value. An approximate mode can be computed by setting the
    number of first unique values to count.

    Parameters:
        k (int): Only the first ``k`` unique values will be included. If ``k`` equals -1, the exact
            mode is computed.

    Attributes:
        counts (collections.defaultdict): Counts the number of occurrences of the different keys.

    Example:

        ::

            >>> from creme import stats

            >>> X = ['sunny', 'cloudy', 'cloudy', 'rainy', 'rainy', 'rainy']
            >>> mode = stats.Mode(k=2)
            >>> for x in X:
            ...     print(mode.update(x).get())
            sunny
            sunny
            cloudy
            cloudy
            cloudy
            cloudy

            >>> mode = stats.Mode(k=-1)
            >>> for x in X:
            ...     print(mode.update(x).get())
            sunny
            sunny
            cloudy
            cloudy
            cloudy
            rainy

    """

    def __init__(self, k=25):
        self.k = k
        self.counts = collections.defaultdict(int)

    @property
    def name(self):
        return 'mode'

    def update(self, x):
        if self.k == -1 or x in self.counts or len(self.counts) < self.k:
            self.counts[x] += 1
        return self

    def get(self):
        return max(self.counts, key=self.counts.get)



"""
A set of metrics used in machine learning that can be computed in a streaming fashion, without any
loss in precision.
"""
from .accuracy import Accuracy
from .confusion import ConfusionMatrix
from .cross_entropy import CrossEntropy
from .f1_score import F1Score
from .f1_score import MacroF1Score
from .f1_score import MicroF1Score
from .log_loss import LogLoss
from .mae import MAE
from .mse import MSE
from .precision import MacroPrecision
from .precision import MicroPrecision
from .precision import Precision
from .recall import MacroRecall
from .recall import MicroRecall
from .recall import Recall
from .rmse import RMSE
from .rmsle import RMSLE
from .smape import SMAPE


__all__ = [
    'Accuracy',
    'ConfusionMatrix',
    'CrossEntropy',
    'F1Score',
    'LogLoss',
    'MacroF1Score',
    'MacroPrecision',
    'MacroRecall',
    'MicroF1Score',
    'MicroPrecision',
    'MicroRecall',
    'MAE',
    'MSE',
    'Precision',
    'Recall',
    'RMSE',
    'RMSLE',
    'SMAPE'
]



import abc
import copy


class Distribution(abc.ABC):

    @abc.abstractmethod
    def update(self, x):
        """Updates the parameters of the distribution given a new observation."""

    @abc.abstractmethod
    def mode(self):
        """Returns the value with the largest likelihood."""

    @abc.abstractmethod
    def __iadd__(self, other):
        """Merges with another distribution or a constant, in-place."""

    @abc.abstractmethod
    def __imul__(self, constant):
        """Scales a distribution by a constant, in-place."""

    def __add__(self, other):
        clone = copy.copy(self)
        clone += other
        return clone

    def __mul__(self, constant):
        clone = copy.copy(self)
        clone *= constant
        return clone

    def __rmul__(self, constant):
        return self * constant


class ContinuousDistribution(Distribution):

    @abc.abstractmethod
    def pdf(self, x):
        """Probability density function."""


class DiscreteDistribution(Distribution):

    @abc.abstractmethod
    def pmf(self, x):
        """Probability mass function."""



import bisect
import collections


class Window(collections.deque):
    """Running window data structure.

    This is just a convenience layer on top of a `collections.deque`.

    Parameters:
        window_size (int): Size of the rolling window.

    Example:

        ::

            >>> from creme import utils

            >>> window = utils.Window(window_size=2)

            >>> for x in [1, 2, 3, 4, 5, 6]:
            ...     print(window.append(x))
            deque([1], maxlen=2)
            deque([1, 2], maxlen=2)
            deque([2, 3], maxlen=2)
            deque([3, 4], maxlen=2)
            deque([4, 5], maxlen=2)
            deque([5, 6], maxlen=2)

    """

    def __init__(self, window_size):
        super().__init__([], maxlen=window_size)

    @property
    def window_size(self):
        return self.maxlen

    def append(self, x):
        super().append(x)
        return self


class SortedWindow(collections.UserList):
    """Sorted running window data structure.

    Parameters:
        window_size (int): size of the window to compute the rolling quantile.

    Example:

        ::

            >>> from creme import utils

            >>> window = utils.SortedWindow(window_size=3)

            >>> for i in reversed(range(9)):
            ...     print(window.append(i))
            [8]
            [7, 8]
            [6, 7, 8]
            [5, 6, 7]
            [4, 5, 6]
            [3, 4, 5]
            [2, 3, 4]
            [1, 2, 3]
            [0, 1, 2]

    References:

        1. `Left sorted <https://stackoverflow.com/questions/8024571/insert-an-item-into-sorted-list-in-python>`_

    """

    def __init__(self, window_size):
        super().__init__()
        self.unsorted_window = Window(window_size)

    @property
    def window_size(self):
        return self.unsorted_window.window_size

    def append(self, x):

        if len(self) >= self.unsorted_window.maxlen:
            self.remove(self.unsorted_window[0])

        bisect.insort_left(self, x)
        self.unsorted_window.append(x)

        return self



from . import moments


class Kurtosis(moments.CentralMoments):
    """Running kurtosis using Welford's algorithm.

    Parameters:
        bias (bool): If ``False``, then the calculations are corrected for statistical bias.

    Attributes:
        centralmoments (stats.CentralMoments)

    Example:

        ::

            >>> import creme.stats
            >>> import scipy.stats
            >>> import numpy as np

            >>> np.random.seed(42)
            >>> X = np.random.normal(loc=0, scale=1, size=10)

            >>> kurtosis = creme.stats.Kurtosis(bias=False)
            >>> for x in X:
            ...     print(kurtosis.update(x).get())
            -3
            -2.0
            -1.5
            1.4130027920707047
            0.15367976585756438
            0.46142633246812653
            -1.620647789230658
            -1.3540178492487054
            -1.2310268787102745
            -0.9490372374384453

            >>> for i in range(1, len(X)+1):
            ...     print(scipy.stats.kurtosis(X[:i], bias=False))
            -3.0
            -2.0
            -1.4999999999999998
            1.4130027920707082
            0.15367976585756082
            0.46142633246812403
            -1.620647789230658
            -1.3540178492487063
            -1.2310268787102738
            -0.9490372374384459

            >>> kurtosis = creme.stats.Kurtosis(bias=True)
            >>> for x in X:
            ...     print(kurtosis.update(x).get())
            -3
            -2.0
            -1.5
            -1.011599627723906
            -0.9615800585356089
            -0.6989395431537853
            -1.4252699121794408
            -1.311437071070812
            -1.246289111322894
            -1.082283689864171

            >>> for i in range(1, len(X)+1):
            ...     print(scipy.stats.kurtosis(X[:i], bias=True))
            -3.0
            -2.0
            -1.4999999999999998
            -1.0115996277239057
            -0.9615800585356098
            -0.6989395431537861
            -1.425269912179441
            -1.3114370710708125
            -1.2462891113228936
            -1.0822836898641714

    References:

        1. `Welford's online algorithm <https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Welford's_Online_algorithm>`_

    """

    def __init__(self, bias=False):
        super().__init__()
        self.bias = bias

    @property
    def name(self):
        return 'kurtosis'

    def update(self, x):
        self.count.update()
        self._update_delta(x)
        self._update_m1(x)
        self._update_sum_delta()
        self._update_m4()
        self._update_m3()
        self._update_m2()
        return self

    def get(self):
        n = self.count.get()
        kurtosis = n * self.M4 / self.M2 ** 2 if self.M2 != 0 else 0
        if not self.bias and n > 3:
            return 1 / (n - 2) / (n - 3) * ((n ** 2 - 1) * kurtosis - 3 * (n - 1) ** 2)
        return kurtosis - 3



import collections

from .. import utils

from . import base


class RollingMode(base.Univariate, utils.Window):
    """Running mode over a window.

    The mode is the most common value.

    Parameters:
        window_size (int): Size of the rolling window.

    Attributes:
        counts (collections.defaultdict): Value counts.

    Example:

        ::

            >>> from creme import stats

            >>> X = ['sunny', 'sunny', 'sunny', 'rainy', 'rainy', 'rainy', 'rainy']
            >>> rolling_mode = stats.rolling_mode.RollingMode(window_size=2)
            >>> for x in X:
            ...     print(rolling_mode.update(x).get())
            sunny
            sunny
            sunny
            sunny
            rainy
            rainy
            rainy

            >>> rolling_mode = stats.rolling_mode.RollingMode(window_size=5)
            >>> for x in X:
            ...     print(rolling_mode.update(x).get())
            sunny
            sunny
            sunny
            sunny
            sunny
            rainy
            rainy

    """

    def __init__(self, window_size):
        super().__init__(window_size)
        self.counts = collections.defaultdict(int)

    @property
    def name(self):
        return f'rolling_{self.window_size}_mode'

    def update(self, x):
        if len(self) >= self.window_size:

            # Subtract the counter of the last element
            first_in = self[0]
            self.counts[first_in] -= 1

            # No need to store the value if it's counter is 0
            if self.counts[first_in] == 0:
                self.counts.pop(first_in)

        self.counts[x] += 1
        super().append(x)
        return self

    def get(self):
        return max(self.counts, key=self.counts.get)



import abc
import collections
import copy
import functools
import itertools
import math

try:
    import graphviz
    GRAPHVIZ_INSTALLED = True
except ImportError:
    GRAPHVIZ_INSTALLED = False
try:
    from matplotlib import patches
    import matplotlib.pyplot as plt
    MATPLOTLIB_INSTALLED = True
except ImportError:
    MATPLOTLIB_INSTALLED = False
from sklearn import utils

from .. import base
from .. import dist


class Segment:
    """A Segment represents a straight line in space."""

    def __init__(self, lower=math.inf, upper=-math.inf):
        self.lower = lower
        self.upper = upper

    def update(self, x):
        """Updates the lower and upper bounds given a new value.

        Example:

            >>> segment = Segment()
            >>> segment
            [inf, -inf]

            >>> segment.update(-3)
            [-3, -3]

            >>> segment.update(42)
            [-3, 42]

        """
        self.lower = min(self.lower, x)
        self.upper = max(self.upper, x)
        return self

    @property
    def length(self):
        return self.upper - self.lower

    def __repr__(self):
        return f'[{self.lower}, {self.upper}]'

    def __or__(self, other):
        """Returns the union with another ``Segment``.

        Example:

            >>> Segment(-1, 2) | Segment(-2, 1)
            [-2, 2]

        """
        return Segment(min(self.lower, other.lower), max(self.upper, other.upper))

    def __and__(self, other):
        """Returns the intersection with another ``Segment``.

        Example:

            >>> Segment(-1, 2) & Segment(-2, 1)
            [-1, 1]

        """
        return Segment(max(self.lower, other.lower), min(self.upper, other.upper))

    def __lt__(self, x):
        """Returns whether or not a value ``x`` is to the right of the ``Segment``.

        Example:

            >>> Segment(-1, 2) < 3
            True

            >>> Segment(-1, 2) < 0
            False

        """
        return self.upper < x

    def __gt__(self, x):
        """Returns whether or not a value ``x`` is to the left of the ``Segment``.

        Example:

            >>> -2 < Segment(-1, 2)
            True

            >>> 0 < Segment(-1, 2)
            False

        """
        return x < self.lower


class Split:

    def __init__(self, feature, threshold):
        self.feature = feature
        self.threshold = threshold

    def __str__(self):
        return f'{self.feature} < {self.threshold:.3f}'


class Node(abc.ABC):
    """A node is an element of a tree and represents a subspace."""

    def __init__(self, box, split_time, target_dist, n_samples):
        self.box = box
        self.split_time = split_time
        self.target_dist = target_dist
        self.n_samples = n_samples

    @property
    def height(self):
        """Returns the length from here to the furthest leaf."""
        if isinstance(self, Branch):
            return 1 + max(self.left.height, self.right.height)
        return 0

    @property
    def size(self):
        """Returns the number of nodes under here."""
        if isinstance(self, Branch):
            return 1 + self.left.size + self.right.size
        return 1

    def extend(self, x, y, lifetime, parent_split_time, min_samples_split, depth, max_depth,
               random_state):
        """Updates a Node given a new observation."""

        # Step 1: Calculate by how much x is outside of the bounding box
        extent = {}
        e_l, e_u = {}, {}
        total_extent = 0

        for i, xi in x.items():
            e_l[i] = max((self.box[i].lower if i in self.box else xi) - xi, 0)
            e_u[i] = max(xi - (self.box[i].upper if i in self.box else xi), 0)
            extent[i] = e_l[i] + e_u[i]
            total_extent += extent[i]

        # Sample E from an exponential distribution
        E = random_state.exponential(scale=1 / total_extent) if total_extent > 0 else math.inf

        # Determine if a split should happen or not
        if parent_split_time + E < self.split_time and \
           self.n_samples + 1 >= min_samples_split and \
           depth < max_depth:

            # Choose which feature to split on
            names = list(extent.keys())
            weights = [extent[name] / total_extent for name in names]
            feature = random_state.choice(a=names, p=weights)

            # Sample a split threshold
            if x[feature] < self.box[feature]:
                threshold = random_state.uniform(low=x[feature], high=self.box[feature].lower)
            else:
                threshold = random_state.uniform(low=self.box[feature].upper, high=x[feature])

            # Create a new leaf containing the current sample
            new_leaf_box = {i: Segment(xi, xi) for i, xi in x.items()}
            new_leaf = Leaf(
                box=new_leaf_box,
                split_time=lifetime,
                target_dist=self.target_dist.__class__().update(y),
                n_samples=1
            )

            # Decide how to arrange the two new leaves
            if x[feature] < self.box[feature]:
                left, right = new_leaf, self
            else:
                left, right = self, new_leaf

            return Branch(
                split=Split(feature, threshold),
                left=left,
                right=right,
                box={i: new_leaf_box[i] | self.box[i] for i in x},
                split_time=parent_split_time + E,
                target_dist=copy.copy(self.target_dist).update(y),
                n_samples=self.n_samples + 1
            )

        # Update the limits of the Node's box
        for i, xi in x.items():
            self.box[i].update(xi)

        # Update the distribution of the target
        self.target_dist.update(y)

        # Increment the number of observed samples
        self.n_samples += 1

        # Recurse down the tree
        if isinstance(self, Branch):
            kwargs = dict(
                x=x,
                y=y,
                lifetime=lifetime,
                parent_split_time=self.split_time,
                min_samples_split=min_samples_split,
                depth=depth + 1,
                max_depth=max_depth,
                random_state=random_state
            )
            if x[self.split.feature] < self.split.threshold:
                self.left = self.left.extend(**kwargs)
            else:
                self.right = self.right.extend(**kwargs)

        return self

    def predict(self, x, target_dist, parent_split_time, p_nsy):
        """Passes an observation through the tree and updates the target distribution."""

        split_time_diff = self.split_time - parent_split_time

        total_extent = 0
        for i, xi in x.items():
            total_extent += max(xi - (self.box[i].upper if i in self.box else xi), 0)
            total_extent += max((self.box[i].lower if i in self.box else xi) - xi, 0)

        weight = p_nsy

        if isinstance(self, Branch):
            p_js = 1 - math.exp(-split_time_diff * total_extent)
            weight *= p_js

            kwargs = dict(
                x=x,
                target_dist=target_dist,
                parent_split_time=self.split_time,
                p_nsy=p_nsy * (1 - p_js)
            )
            if x[self.split.feature] < self.split.threshold:
                target_dist = self.left.predict(**kwargs)
            else:
                target_dist = self.right.predict(**kwargs)

        # Update the target distribution using the weight
        target_dist += weight * self.target_dist

        return target_dist

    def to_dot(self):

        if not GRAPHVIZ_INSTALLED:
            raise RuntimeError('graphviz is not installed')

        dot = graphviz.Digraph()

        def add_node(node, code):

            # Draw current node
            if isinstance(node, Leaf):
                dot.node(code, f'y = {node.target_dist.mode()}\n' +
                               f'n_samples = {node.n_samples}\n' +
                               f'split_time = {node.split_time:.3f}')
            else:
                dot.node(code, f'{node.split}\n' +
                               f'y = {node.target_dist.mode()}\n' +
                               f'n_samples = {node.n_samples}\n' +
                               f'split_time = {node.split_time:.3f}')
                add_node(node.left, f'{code}0')
                add_node(node.right, f'{code}1')

            # Draw link with the previous node
            is_root = len(code) == 1
            if not is_root:
                dot.edge(code[:-1], code)

        add_node(self, '0')

        return dot

    def plot(self, x=None, y=None, ax=None, colors=None, alpha=1):

        if not MATPLOTLIB_INSTALLED:
            raise RuntimeError('matplotlib is not installed')

        if x is None and y is None:
            raise ValueError('At least one of x and y must be provided')

        ax = ax or plt.axes()
        colors = colors or plt.rcParams['axes.prop_cycle'].by_key()['color']
        colors = itertools.cycle(colors)

        def plot_node(node):

            if isinstance(node, Branch):
                plot_node(node.left)
                plot_node(node.right)
                return

            x_segment = node.box.get(x)
            y_segment = node.box.get(y)
            style = dict(color=next(colors), alpha=alpha)

            # If both the x-axis and the y-axis are given then draw a box
            if x_segment is not None and y_segment is not None:
                ax.add_patch(patches.Rectangle(
                    xy=(x_segment.lower, y_segment.lower),
                    width=x_segment.length,
                    height=y_segment.length,
                    **style
                ))
                return

            # If only the x-axis is given then draw a vertical band
            if x_segment is not None:
                ax.axvspan(xmin=x_segment.lower, xmax=x_segment.upper, **style)

            # If only the y-axis is given then draw a horizontal line
            if y_segment is not None:
                ax.axhspan(ymin=y_segment.lower, ymax=y_segment.upper, **style)
                return

        plot_node(self)
        ax.autoscale_view()

        return ax


class Branch(Node):
    """A Branch is a Node that has two children that are also Nodes."""

    def __init__(self, split, left, right, **kwargs):
        super().__init__(**kwargs)
        self.split = split
        self.left = left
        self.right = right


class Leaf(Node):
    """A Leaf is a Node that has no children."""


class MondrianTree:

    def __init__(self, target_dist, lifetime, max_depth, min_samples_split, random_state):
        self.tree = Leaf(
            box=collections.defaultdict(functools.partial(Segment, math.inf, -math.inf)),
            split_time=lifetime,
            target_dist=copy.copy(target_dist),
            n_samples=0
        )
        self.lifetime = lifetime
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.random_state = utils.check_random_state(random_state)

    def fit_one(self, x, y):

        # Start by getting the prediction from the tree
        y_pred = self.predict_one(x)

        # Update the tree
        self.tree = self.tree.extend(
            x=x,
            y=y,
            lifetime=self.lifetime,
            parent_split_time=0,
            min_samples_split=self.min_samples_split,
            depth=0,
            max_depth=self.max_depth,
            random_state=self.random_state
        )

        return y_pred

    def predict_one(self, x):
        target_dist = self.tree.predict(
            x=x,
            target_dist=self.tree.target_dist.__class__(),
            parent_split_time=0,
            p_nsy=1
        )
        return target_dist.mode()


class MondrianTreeRegressor(MondrianTree, base.Regressor):
    """Mondrian tree for regression.

    Example:

        >>> from creme import compose
        >>> from creme import metrics
        >>> from creme import model_selection
        >>> from creme import optim
        >>> from creme import preprocessing
        >>> from creme import stream
        >>> from creme import tree
        >>> from sklearn import datasets

        >>> X_y = stream.iter_sklearn_dataset(
        ...     load_dataset=datasets.load_boston,
        ...     shuffle=True,
        ...     random_state=42
        ... )
        >>> model = compose.Pipeline([
        ...     ('scale', preprocessing.StandardScaler()),
        ...     ('mondrian', tree.MondrianTreeRegressor(random_state=42))
        ... ])
        >>> metric = metrics.MAE()

        >>> model_selection.online_score(X_y, model, metric)
        MAE: 4.250497

    References:

    1. `Mondrian Forests: Efficient Online Random Forests <https://github.com/balajiln/mondrianforest>`_
    2. `Decision Trees and Forests: A Probabilistic Perspective <http://www.gatsby.ucl.ac.uk/~balaji/balaji-phd-thesis.pdf>`_
    3. `scikit-garden <https://github.com/scikit-garden/scikit-garden/tree/master/skgarden/mondrian>`_

    """

    def __init__(self, lifetime=1, max_depth=8, min_samples_split=1, random_state=None):
        return super().__init__(
            target_dist=dist.Normal(),
            lifetime=lifetime,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            random_state=random_state
        )


class MondrianTreeClassifier(MondrianTree, base.MultiClassifier):
    """Mondrian tree for classification.

    Example:

        ::

            >>> from creme import compose
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from creme import tree
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     load_dataset=datasets.load_breast_cancer,
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> model = compose.Pipeline([
            ...     ('scale', preprocessing.StandardScaler()),
            ...     ('mondrian', tree.MondrianTreeClassifier(random_state=42))
            ... ])
            >>> metric = metrics.F1Score()

            >>> model_selection.online_score(X_y, model, metric)
            F1Score: 0.823377

    References:

        1. `Mondrian Forests: Efficient Online Random Forests <https://github.com/balajiln/mondrianforest>`_
        2. `Decision Trees and Forests: A Probabilistic Perspective <http://www.gatsby.ucl.ac.uk/~balaji/balaji-phd-thesis.pdf>`_
        3. `scikit-garden <https://github.com/scikit-garden/scikit-garden/tree/master/skgarden/mondrian>`_

    """

    def __init__(self, lifetime=1, max_depth=5, min_samples_split=1, random_state=None):
        return super().__init__(
            target_dist=dist.Multinomial(),
            lifetime=lifetime,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            random_state=random_state
        )

    def predict_proba_one(self, x):
        target_dist = self.tree.predict(
            x=x,
            target_dist=self.tree.target_dist.__class__(),
            parent_split_time=0,
            p_nsy=1
        )
        y_pred = target_dist.counter
        y_pred_sum = sum(y_pred.values())
        return {c: y / y_pred_sum for c, y in y_pred.items()}



from .. import stats

from . import base


__all__ = ['SMAPE']


class SMAPE(stats.Mean, base.RegressionMetric):
    """Symmetric mean absolute percentage error.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [100, 100]
            >>> y_pred = [110, 90]

            >>> metric = metrics.SMAPE()
            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(metric.update(y_t, y_p))
            SMAPE: 4.761905
            SMAPE: 5.012531

            >>> metric
            SMAPE: 5.012531

    """

    @property
    def bigger_is_better(self):
        return False

    def update(self, y_true, y_pred):
        return super().update(abs(y_true - y_pred) / (abs(y_true) + abs(y_pred)))

    def get(self):
        return 100 * super().get()
